"""
This module provides variations of linear layers, including linear layers that output
multiple tensors and a linear layer with a weight tied to another module's parameter.
"""

import itertools as _itertools
import logging as _logging
from typing import List as _List
from typing import Optional as _Optional

import torch as _torch
import torch.nn as _nn
from torch.nn import functional as _F

from tamm.layers import functional as _tamm_F
from tamm.layers import init as _init
from tamm.layers import sequential as _sequential
from tamm.layers.common import LayerMixin as _LayerMixin
from tamm.utils import torch_utils as _torch_utils

_logger = _logging.getLogger(__name__)


class Linear(_nn.Linear, _LayerMixin):
    """
    Embedding layer with variance scaled init.
    """

    def forward(self, *args, **kwargs):
        # pylint: disable=W0246
        return super().forward(*args, **kwargs)

    def reset_parameters(self) -> None:
        _init.shape_normalized_normal_(self.weight, dim=1)
        if self.bias is not None:
            _nn.init.zeros_(self.bias)


class MultiOutputLinear(_nn.Module, _LayerMixin):
    """
    A linear layer that outputs multiple tensors given the same input.

    Args:
        input_dim (:obj:`int`): The input's dimension along its last axis.
        output_dims (:obj:`int`): A list of one or more output dimensions.
        bias (:obj:`bool`): Will the linear layers use a bias component or not.
    """

    def __init__(
        self,
        input_dim: int,
        output_dims: _List[int],
        bias: bool = False,
        *,
        vec_dim: _Optional[_List[int]] = None,
        device=None,
        dtype=None,
        override_linear_cls=None,
    ):
        super().__init__()
        self.output_dims = output_dims
        self.bias = bias

        # Handle vectorized
        self.vec_dim = vec_dim
        if vec_dim is None:
            linear_cls = Linear
            extra_kwargs = {}
        else:
            linear_cls = VectorizedLinear
            extra_kwargs = {"vec_dim": vec_dim}

        # Option for overriding the linear cls
        if override_linear_cls is not None:
            linear_cls = override_linear_cls

        for idx, output_dim in enumerate(output_dims):
            linear = linear_cls(
                input_dim,
                output_dim,
                bias=self.bias,
                device=device,
                dtype=dtype,
                **extra_kwargs,
            )
            self.register_module(f"linear_{idx}", linear)

    # pylint: disable-next=all
    def forward(self, input, **kwargs):
        return tuple(linear(input, **kwargs) for linear in self.children())

    def extra_repr(self) -> str:
        return f"out_features={self.output_dims}, bias={self.bias}"

    def _load_from_state_dict(self, state_dict, prefix, *args, **kwargs):
        """
        MultiOutputLinear and FusedMultiOutputLinear are interchangeable except for
        the state_dict format.  It is useful for users to have the option of swapping
        them because some adapters support one and not the other.  Here we override
        nn.Module._load_from_state_dict so that MultiOutputLinear can load state
        dicts from FusedMultiOutputLinear.  This needs to rely on PyTorch internals
        because PyTorch does not otherwise provide a way of extending load_state_dict.
        """
        self._adapt_state_dict_from_fused_format(state_dict, prefix=prefix)
        return super()._load_from_state_dict(state_dict, prefix, *args, **kwargs)

    def _adapt_state_dict_from_fused_format(self, state_dict, prefix=""):
        param_names = ["weight"]
        if self.bias:
            param_names.append("bias")

        for param_name in param_names:
            fused_param_key = f"{prefix}fused_linear.{param_name}"

            if fused_param_key not in state_dict:  # nothing to do for this param
                continue

            fused_param = state_dict[fused_param_key]

            if param_name == "weight":
                expected_shape = (sum(self.output_dims), self.linear_0.in_features)
            else:
                expected_shape = (sum(self.output_dims),)

            # stop processing if shape's off
            if fused_param.shape != expected_shape:
                return

            new_param_keys = [
                f"{prefix}linear_{idx}.{param_name}"
                for idx, _ in enumerate(self.output_dims)
            ]
            # Stop processing if any of the un-fused tensors exist in the state_dict
            for key in new_param_keys:
                if key in state_dict:
                    return

            _logger.debug(
                f"Mapping FusedMultiOutputLinear's {param_name} "
                f"to MultiOutputLinear format"
            )
            state_dict.pop(fused_param_key)
            split_weights = _torch.split(fused_param, self.output_dims)
            for key, weight in zip(new_param_keys, split_weights):
                state_dict[key] = weight


class FusedMultiOutputLinear(_nn.Module, _LayerMixin):
    """
    A :class:`nn.Module` for fusing multiple linear layers (with the same input dim) as
    a single linear layer.

    Args:
        input_dim (:obj:`int`): The input's dimension along its last axis.
        output_dims (:obj:`int`): A list of one or more output dimensions.
        bias (:obj:`bool`): Whether the linear layer uses a bias or not.
    """

    def __init__(
        self,
        input_dim: int,
        output_dims: _List[int],
        bias: bool = False,
        *,
        vec_dim: _Optional[_List[int]] = None,
        device=None,
        dtype=None,
    ):
        super().__init__()
        self.input_dim = input_dim
        self.output_dims = output_dims
        self.bias = bias

        # Handle vectorized
        self.vec_dim = vec_dim
        if vec_dim is None:
            linear_cls = Linear
            extra_kwargs = {}
        else:
            linear_cls = VectorizedLinear
            extra_kwargs = {"vec_dim": vec_dim}

        self.fused_linear = linear_cls(
            input_dim,
            sum(output_dims),
            bias=bias,
            device=device,
            dtype=dtype,
            **extra_kwargs,
        )

    # pylint: disable-next=all
    def forward(self, inp: _torch.Tensor):
        output = self.fused_linear(inp)
        return _torch.split(output, self.output_dims, dim=-1)

    def extra_repr(self) -> str:
        return f"out_features={self.output_dims}, bias={self.bias}"

    def _load_from_state_dict(self, state_dict, prefix, *args, **kwargs):
        """
        MultiOutputLinear and FusedMultiOutputLinear are interchangeable except for
        the state_dict format.  It is useful for users to have the option of swapping
        them because some adapters support one and not the other.  Here we override
        nn.Module._load_from_state_dict so that FusedMultiOutputLinear can load state
        dicts from MultiOutputLinear.  This needs to rely on PyTorch internals because
        PyTorch does not otherwise provide a way of extending load_state_dict.
        """
        self.adapt_state_dict_from_unfused_format(
            state_dict, bias=self.bias, prefix=prefix
        )
        return super()._load_from_state_dict(state_dict, prefix, *args, **kwargs)

    @staticmethod
    def adapt_state_dict_from_unfused_format(state_dict, bias, prefix=""):
        weight_names = ["weight"]
        if bias:
            weight_names.append("bias")
        for weight_name in weight_names:
            matching_keys = []
            for idx in _itertools.count():
                key = f"{prefix}linear_{idx}.{weight_name}"
                if key not in state_dict:
                    break
                matching_keys.append(key)

            if len(matching_keys) > 0:
                _logger.debug(
                    f"Mapping MultiOutputLinear's {weight_name} "
                    f"to FusedMultiOutputLinear format"
                )
                tensors = [state_dict.pop(key) for key in matching_keys]
                fused_key = f"{prefix}fused_linear.{weight_name}"
                assert fused_key not in state_dict
                state_dict[f"{prefix}fused_linear.{weight_name}"] = _torch.cat(tensors)


class TiedWeightLinear(_nn.Module, _LayerMixin):
    """
    A replacement for :class:`nn.Linear` that uses a weight parameter from another
    :obj:`nn.Module`.

    Args:
        module (:obj:`nn.Module`): The module with the weight parameter.  The weight
            parameter should have shape ``(out_features, in_features)``.
        parameter_name (:obj:`str`): The name of the weight parameter registered on
            ``module``.  (The parameter must be accessible via
            ``getattr(module, parameter_name)``.)
    """

    _WEIGHT_NAME = "weight"

    def __init__(
        self,
        module: _nn.Module,
        parameter_name: str,
        transpose: bool = False,
    ):
        super().__init__()
        self._parameters = _torch_utils.TieableParamsDict(self._parameters)

        # Implementation note: We access the tied weight as an attribute of another
        # module rather than simply as a nn.Parameter due to FSDP behavior as of
        # Torch 2.0.  Specifically, FSDP replaces parameters on modules with sharded
        # parameters. Thus, parameters would become untied between modules unless we can
        # still access the new sharded parameter via the module's updated attribute.

        self._parameters.register_tied_parameter(
            name=self._WEIGHT_NAME,
            tied_module=module,
            tied_param_name=parameter_name,
            # This ensures that every time we access the param (for example
            # layer.state_dict()) we have the updated value from the tied module.
        )

        if len(self._weight_shape) != 2:
            raise RuntimeError(
                "TiedWeightLinear received a weight tensor with dim "
                f"{len(self._weight_shape)} (expected 2)"
            )

        self.transpose = transpose

        if self.transpose:
            self.in_features, self.out_features = self._weight_shape
        else:
            self.out_features, self.in_features = self._weight_shape

        self._maybe_register_deepspeed_external_param()

    @property
    def weight(self):
        # This property is almost not needed because nn.Module.__getattr__() will already
        # return the weight from _parameters.  However, this only happens when the weight
        # is a nn.Parameter, and FSDP sometimes replaces it with a vanilla torch.Tensor.
        # Adding this property ensures it still works when weight becomes a non-Param.
        return self._parameters.get_tied_value(self._WEIGHT_NAME)

    @property
    def _weight_shape(self):
        if hasattr(self.weight, "ds_shape"):  # special case to support deepspeed
            return self.weight.ds_shape
        return self.weight.shape

    def _maybe_register_deepspeed_external_param(self):
        if not hasattr(self.weight, "ds_id"):
            return
        try:
            from deepspeed import zero  # pylint: disable=all
        except Exception:
            _logger.debug(
                "TiedWeightLinear failed to import deepspeed to register its external param"
            )
            return

        zero.register_external_parameter(self, self.weight)
        _logger.debug(
            "TiedWeightLinear registered its weight as a deepspeed external param"
        )

    def __setattr__(self, name, value):
        if name == self._WEIGHT_NAME:
            # As of torch 2.5, FSDP calls this to replace the param with a tensor.
            # We want to just ignore it and still use the param of the tied module.
            _logger.debug("Ignoring call to set weight of TiedWeightLinear layer")
            return
        super().__setattr__(name, value)

    def __delattr__(self, name):
        if name == self._WEIGHT_NAME:
            # As of torch 2.5, FSDP calls this to replace the param with a tensor.
            # We want to just ignore it and still use the param of the tied module.
            _logger.debug("Ignoring call to delete weight of TiedWeightLinear layer")
            return
        super().__delattr__(name)

    # pylint: disable-next=all
    def forward(self, inp: _torch.Tensor) -> _torch.Tensor:
        # As of torch 2.5, using self.weight here causes issues with torch.compile
        weight = self._parameters.get_tied_value(self._WEIGHT_NAME)
        if self.transpose:
            weight = weight.transpose(-1, -2)
        return _F.linear(inp, weight)  # pylint: disable=not-callable

    def reset_parameters(self):
        pass

    def extra_repr(self) -> str:
        return f"in_features={self.in_features}, out_features={self.out_features}"


class TiedWeightLinearSequence(_sequential.Sequential):
    """
    A sequence of :class:`TiedWeightLinear` layers that share the same ``module``.

    Args:
        module (:obj:`nn.Module`):
    """

    def __init__(
        self,
        module: _nn.Module,
        parameter_names: _List[str],
        transpose_flags: _Optional[_List[bool]] = None,
    ):
        if transpose_flags is None:
            transpose_flags = [False] * len(parameter_names)

        named_layers = {}
        for idx, (name, transpose) in enumerate(zip(parameter_names, transpose_flags)):
            layer = TiedWeightLinear(
                module=module, parameter_name=name, transpose=transpose
            )
            named_layers[f"layer_{idx}"] = layer

        super().__init__(named_layers)


# Implementation of _vectorized_ Linear operations.
class VectorizedLinear(_LayerMixin, _nn.Module):
    def __init__(
        self,
        in_features: int,
        out_features: int,
        vec_dim: _List[int] = (1,),
        bias: bool = False,
        *,
        device=None,
        dtype=None,
    ):
        # Same as Linear but vec_dim is a tuple/list of ints
        # that replicate the linear layer along these dimensions.
        #
        # Example: VectorizedLinear(3, 5, vec_dim=[1,8])
        # -> creates linear layer with weights of shape (1,8,3)
        # -> operator on `input` does input @ weights (bmm)
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.vec_dim = vec_dim

        weight_shape = tuple(vec_dim) + (in_features, out_features)
        self.weight = _nn.Parameter(_torch.empty(weight_shape, **factory_kwargs))

        if bias:
            bias_shape = tuple(vec_dim) + (out_features,)
            self.bias = _nn.Parameter(_torch.empty(bias_shape, **factory_kwargs))
        else:
            self.register_parameter("bias", None)

        self.reset_parameters()

    def reset_parameters(self) -> None:
        _init.shape_normalized_normal_(self.weight, dim=-2)
        if self.bias is not None:
            _nn.init.zeros_(self.bias)

    # pylint: disable=redefined-builtin
    def forward(self, input: _torch.Tensor) -> _torch.Tensor:
        # Optimize decode performance when batch > seqlen for ParallelTrack models.
        #
        # Specifically target input shapes of format [batch, tracks, seqlen, input_dim]
        # where weight shapes are of format [tracks, input_dim, output_dim]
        #
        # During decode, seqlen=1, while batch can be high; in this case, it's much faster
        # to swap batch and seqlen dimensions so that matrix-matrix multiplications are used.
        optimize_decode_inference = (
            len(input.shape) == 4
            and len(self.weight.shape) == 3
            and input.shape[1] == self.weight.shape[0]
            and input.shape[0] > input.shape[2]
        )
        if optimize_decode_inference:
            input = _torch.swapaxes(input, 0, 2)

        x = _torch.matmul(input, self.weight)

        if optimize_decode_inference:
            x = _torch.swapaxes(x, 0, 2)

        if self.bias is not None:
            x = x + self.bias
        return x

    def extra_repr(self) -> str:
        return (
            f"in_features={self.in_features}, "
            f"out_features={self.out_features}, "
            f"vec_dim={self.vec_dim}, bias={self.bias is not None}"
        )


# Implementation of segmented linear operations. See functiona segment_matmul for details.
class SegmentedLinear(VectorizedLinear):
    # pylint: disable=redefined-builtin
    def forward(self, input: _torch.Tensor, **kwargs) -> _torch.Tensor:
        return _tamm_F.segment_matmul(input, kwargs["group_sizes"], self.weight)



import contextlib
import copy
import dataclasses
import enum as enum_module
import functools
import inspect
import itertools
import logging
import operator
import os
import pathlib
import shutil
import textwrap
import time
import weakref
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Collection,
    ContextManager,
    Dict,
    Iterable,
    List,
    Optional,
    Union,
    cast,
)

import filelock
import torch
import torch.nn as _nn
from packaging.version import parse as parse_version

if TYPE_CHECKING:
    from tamm.typing import (
        OptionalDeviceOrString,
        OptionalDtypeOrString,
        PathLike,
        StateDictType,
    )
logger = logging.getLogger(__name__)


def case_insensitive_lookup(enum_class, value):
    value = value.lower()
    for member in enum_class:
        if member.lower() == value:
            return member
    return None


def get_item_by_index(iterable: "Iterable[Any]", idx: int) -> Any:
    """
    Returns the idx-th item of an iterable.  This function also supports negative
    indexing if the iterable has a __len__() method.

    Example: Get the last key in a dictionary d: get_item_by_index(d.keys(), -1)
    """
    idx = operator.index(idx)
    if idx > 0:
        new_idx = idx
    elif idx >= -len(iterable):
        new_idx = idx % len(iterable)
    else:
        raise IndexError(f"Index {idx} out of range")

    try:
        return next(itertools.islice(iterable, new_idx, None))
    except StopIteration:
        # pylint: disable=raise-missing-from
        raise IndexError(f"Index {idx} out of range")


def count_iterable(itr: Iterable[Any]) -> int:
    return sum(1 for el in itr)


def maybe_slice_to_string(obj):
    if not isinstance(obj, slice):
        return str(obj)
    start = str(obj.start) if obj.start is not None else ""
    stop = str(obj.stop) if obj.stop is not None else ""
    pieces = [start, stop]
    if obj.step is not None:
        pieces.append(str(obj.step))
    return ":".join(pieces)


# pylint: disable-next=inconsistent-return-statements
def make_key_unique(key: str, existing_keys: Collection[str]) -> str:
    """
    Returns the key if it is not in existing_keys.  Otherwise finds a new key that is
    not in existing_keys and returns that.
    """
    for attempt in itertools.count():
        candidate = key if attempt == 0 else f"{key}_{attempt}"
        if candidate not in existing_keys:
            return candidate


def copy_if_not_none(obj: Any) -> Any:
    """Returns None if the object is None and otherwise a shallow copy of the object."""
    return None if obj is None else copy.copy(obj)


def cumsum(numbers: Iterable[Union[int, float]]) -> List[Union[int, float]]:
    """Returns the cumulative sum of an iterable."""
    result = [0] * len(numbers)
    for idx, num in enumerate(numbers):
        result[idx] = result[idx - 1] + num
    return result


def get_dict_key_from_value(dictionary, value):
    for key, keys_value in dictionary.items():
        if keys_value == value:
            return key
    return None


def merge_dicts(*dicts):
    """
    Merges dictionaries into a single dictionary. If a key appears multiple times in
    *dicts, the resulting value comes from the last dict with that key.
    """
    return dict(itertools.chain.from_iterable(d.items() for d in dicts))


def get_enum_member_from_name(
    enum: enum_module.EnumMeta, member_or_name: Union[enum_module.Enum, str]
) -> enum_module.Enum:
    """
    Helper to lookup Enums by [].
    Handle string enums in a special way for torch.compile().

    Context:

        Enum lookup by instantiation, i.e., ``ColorEnum("red")``, generally works in
        Python. However, if ``forward()`` implementation of compiled module involves
        Enum resolution, it will fail ``torch.compile()`` with the following error:

            'Enum variable is constructed with non-constant values'

    Args:
        enum: Enum type (i.e., EnumMeta)
        member_or_name: string or Enum

    Returns: Enum member
    """

    if isinstance(member_or_name, enum):
        return cast(enum_module.Enum, member_or_name)

    if isinstance(member_or_name, str):
        name = member_or_name.upper()
        try:
            return enum[name]  # do not change this, see docstring
        except KeyError as exception:
            valid_choices = [el.name for el in enum]  # type: ignore
            raise ValueError(
                f"Name '{name}' not found in {enum}. "
                f"Valid choices include {valid_choices}."
            ) from exception
    raise TypeError(
        f"Please pass a member of {enum} or a string corresponding to the name of one "
        "of its members."
    )


def maybe_build_module(
    obj: Union[Callable[[], _nn.Module], _nn.Module, None], *args, **kwargs
) -> Union[_nn.Module, None]:
    """
    Helper function for creating layers from objects that may be a LayerBuilder or
    possibly None or a nn.Module already.

    Args:
        obj: Any callable that returns a nn.Module -- or a nn.Module or None.
        args, kwargs: Optional args and kwargs for the callable.

    Returns:
        The nn.Module returned by the callable -- or the existing nn.Module or None if
        the input is already one of those.
    """
    return maybe_build_object(obj, _nn.Module, *args, **kwargs)


def maybe_build_object(obj, target_cls, *args, **kwargs):
    if obj is None:
        return None
    if isinstance(obj, target_cls):
        return obj
    return obj(*args, **kwargs)


def append_children(module, /, **child_names_and_modules):
    """
    Appends new nn.Modules to the end of the list of module's children.

    Args:
        module (:obj:`nn.Module`): The parent module that will receive the children.
        child_names_and_modules: Names and LayerBuilders for the children.
    """
    _assert_module_does_not_have_attributes(module, child_names_and_modules.keys())
    for name, child in child_names_and_modules.items():
        child = maybe_build_module(child)
        if hasattr(module, name):
            delattr(module, name)
        setattr(module, name, child)


def prepend_children(module, /, **child_names_and_modules):
    """
    The same as :func:`append_children` but adds the new children at the start of the
    list.  Adding at the start is sometimes helpful for organizing the module's repr.
    """
    _assert_module_does_not_have_attributes(module, child_names_and_modules.keys())
    child_names_and_modules.update(module.named_children())
    for name, child in child_names_and_modules.items():
        child = maybe_build_module(child)
        if hasattr(module, name):
            delattr(module, name)
        setattr(module, name, child)


def get_all_named_children(module):
    """
    Returns the names of all a module's children, not just the unique children from
    module.named_children().
    """
    for name, submodule in module.named_modules(remove_duplicate=False):
        if name != "" and "." not in name:
            yield name, submodule


def _assert_module_does_not_have_attributes(module, attribute_names):
    for name in attribute_names:
        if hasattr(module, name):
            raise RuntimeError(
                f"{module.__class__.__name__} already has an attribute named {name}"
            )


def are_views_identical(view1, view2):
    if view1.data_ptr() != view2.data_ptr():
        return False
    if view1.shape != view2.shape:
        return False
    if view1.stride() != view2.stride():
        return False
    return True


def get_dtype_from_maybe_string(obj: Union[str, torch.dtype, None]) -> torch.dtype:
    if obj is None:
        return None
    if isinstance(obj, str):
        return getattr(torch, obj)
    return obj


def get_str_from_maybe_dtype(obj: Union[str, torch.dtype, None]) -> Union[str, None]:
    if obj is None:
        return None
    if isinstance(obj, torch.dtype):
        return str(obj)[6:]  # 6: removes the torch. prefix
    return obj


def get_dtype_after_autocast(dtype: torch.dtype, device: torch.device) -> torch.dtype:
    # pylint: disable=import-outside-toplevel
    from tamm.utils import _torch_compatibility

    if not dtype.is_floating_point:
        return dtype
    if _torch_compatibility.is_autocast_enabled(device.type):
        return _torch_compatibility.get_autocast_dtype(device.type)
    return dtype


def autocast_disabled(device: torch.device):
    """Returns a context manager that disabled autocast within the context."""
    if device.type == "mps" and is_torch_base_version_less_than("2.5"):
        return contextlib.nullcontext()  # mps does not support autocast as of torch 2.4
    return torch.autocast(device.type, enabled=False)


def maybe_cast_state_dict_(
    state_dict: "StateDictType", *, dtype: "OptionalDtypeOrString" = None
):
    """
    If dtype is not None, this function casts tensors in state_dict.values() to dtype.
    This modifies the state dict *in place*.
    """
    dtype = get_dtype_from_maybe_string(dtype)
    if dtype is None:
        return
    for key, value in state_dict.items():
        if not torch.is_tensor(value) or not value.is_floating_point():
            continue
        state_dict[key] = value.type(dtype)


def safe_torch_load(*args, **kwargs):
    kwargs["weights_only"] = True
    return torch.load(*args, **kwargs)


def make_noop_when_recursing(fn):
    is_recursing = False

    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        nonlocal is_recursing
        if is_recursing:
            return None
        is_recursing = True
        result = fn(*args, **kwargs)
        is_recursing = False
        return result

    return wrapper


CACHED_FUNCTIONS = weakref.WeakSet()


def cache(user_function: "Callable") -> Callable:
    caching_fn = functools.lru_cache(maxsize=None)(user_function)

    @functools.wraps(caching_fn)
    def wrapper(*args, **kwargs):
        # wrap again because lru_cache does not support weakref in Py3.8
        return caching_fn(*args, **kwargs)

    wrapper.cache_clear = caching_fn.cache_clear

    try:  # register cached functions so that we can reset them during testing
        CACHED_FUNCTIONS.add(wrapper)
    except Exception as e:  # pylint: disable=broad-except
        log_exception(e, context_name="adding weakref to cached function")
    return wrapper


def catch_and_log_exceptions(fn=None, context_name=None):
    if fn is None:
        return _catch_and_log_exceptions_context(context_name=context_name)

    if context_name is None:
        context_name = fn.__name__

    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        context = _catch_and_log_exceptions_context(context_name=context_name)
        with context:
            fn(*args, **kwargs)

    return wrapper


@contextlib.contextmanager
def _catch_and_log_exceptions_context(context_name=None):
    try:
        yield
    except Exception as e:  # pylint: disable=broad-except
        log_exception(e, context_name=context_name)


def log_exception(exception, context_name: Optional[str] = None) -> None:
    name_str = f" during {context_name}" if context_name is not None else ""
    logger.debug("Caught exception%s: %s", name_str, exception, exc_info=True)


def get_all_params_and_buffers(module) -> List[torch.Tensor]:
    return list(itertools.chain(module.parameters(), module.buffers()))


def get_device_context(device: "OptionalDeviceOrString") -> "ContextManager":
    """
    Returns a context manager that changes the default device to ``device``.  If
    ``device`` is None, returns a null context.
    """
    if device is None:
        return contextlib.nullcontext()
    if isinstance(device, torch.device):
        return device
    return torch.device(device)


def move_meta_tensors_to_device(module, device=None):
    """
    Iterates over parameters and buffers in module (and its submodules), overwriting
    them with new tensors on device of the same shape and dtype.  Tied params or buffers
    remain tied.  If device is None, the function uses the default device.
    """
    meta_tensor_to_real_tensor = {}
    for submodule in module.modules():
        named_params_and_buffers = itertools.chain(
            submodule.named_parameters(recurse=False),
            submodule.named_buffers(recurse=False),
        )
        for name, tensor in list(named_params_and_buffers):
            if not tensor.is_meta:
                continue

            if tensor in meta_tensor_to_real_tensor:
                real_tensor = meta_tensor_to_real_tensor[tensor]
            else:
                real_tensor = torch.empty(
                    tensor.shape,
                    dtype=tensor.dtype,
                    device=device,
                    layout=tensor.layout,
                    requires_grad=tensor.requires_grad,
                    # don't use empty_like(tensor) here because if device is None, we
                    # want to use the default device, not tensor's device
                )
                if isinstance(tensor, torch.nn.Parameter):
                    real_tensor = torch.nn.Parameter(
                        real_tensor, requires_grad=tensor.requires_grad
                    )
                meta_tensor_to_real_tensor[tensor] = real_tensor

            if isinstance(real_tensor, torch.nn.Parameter):
                delattr(submodule, name)
                setattr(submodule, name, real_tensor)
            else:
                persistent = name in submodule.state_dict(keep_vars=True)
                delattr(submodule, name)
                submodule.register_buffer(name, real_tensor, persistent=persistent)


def get_timestamp_ms() -> int:
    return time.time_ns() // 1_000_000


@contextlib.contextmanager
def file_lock(lock_path: "PathLike"):
    """
    A simple wrapper around filelock.FileLock that adds logging and creates the parent
    dir for the lock file if it does not already exist.
    """
    lock_path = pathlib.Path(lock_path)
    os.makedirs(lock_path.parent, exist_ok=True)
    logger.debug("Acquiring file lock %s", lock_path)
    with filelock.FileLock(lock_path):
        yield
    logger.debug("File lock released")


def truncate_lines_for_terminal(*lines):
    terminal_size = shutil.get_terminal_size()
    max_width = terminal_size.columns
    return [x if len(x) < max_width else x[: max_width - 1] + "…" for x in lines]


def strip_prefix_from_state_dict_keys(state_dict, prefix):
    new_state_dict = {}
    for key, val in state_dict.items():
        if key.startswith(prefix):
            key = key[len(prefix) :]
        new_state_dict[key] = val
    return new_state_dict


def add_prefix_to_state_dict_keys(state_dict, prefix):
    return {prefix + k: v for k, v in state_dict.items()}


def parse_base_version(version_str: str) -> str:
    """
    Function to parse base version from a given version string.

    Example: version_str of "2.0.1+cu117" provides "2.0.1"
    """
    return parse_version(parse_version(version_str).base_version)


def is_torch_base_version_less_than(version_str: str) -> bool:
    """
    Function to check whether the runtime torch version
    is less than a given base version.
    """
    return parse_base_version(torch.__version__) < parse_base_version(version_str)


def passthrough_decorator(fn):
    """
    This is a decorator that wraps a function but otherwise does nothing.  One use case
    is to create a "copy" of a function for the purpose of setting its __doc__ or
    __signature__ attributes.
    """

    @functools.wraps(fn)
    def wrapped(*args, **kwargs):
        return fn(*args, **kwargs)

    return wrapped


def get_function_from_maybe_method(fn):
    if inspect.ismethod(fn):
        return fn.__func__
    return fn


def add_kw_only_params_to_signature(
    signature: inspect.Signature, kw_only_params: List[inspect.Parameter]
) -> inspect.Signature:
    """
    Creates a new signature, which is the same as ``signature`` but with extra
    keyword-only params from ``kw_only_params``.
    """
    new_params = [
        p for p in signature.parameters.values() if p.kind is not p.VAR_KEYWORD
    ]
    new_params.extend(kw_only_params)
    new_params.extend(
        p for p in signature.parameters.values() if p.kind is p.VAR_KEYWORD
    )
    return signature.replace(parameters=new_params)


def dataclass_to_dict(dataclass, *, omit_defaults: bool = False) -> Dict[str, Any]:
    result = {}
    for field in dataclasses.fields(dataclass):
        value = getattr(dataclass, field.name)
        if omit_defaults and value == field.default:
            continue
        result[field.name] = value
    return result


def set_annotated_class_attr(cls, attr, value, annotation):
    """
    Sets an attribute of cls and also updates cls.__annotations__ with an annotation.
    """
    setattr(cls, attr, value)
    annotations = getattr(cls, "__annotations__", {})
    annotations = copy.copy(annotations)
    annotations[attr] = annotation
    setattr(cls, "__annotations__", annotations)


def maybe_get_enum_member_for_all_str_enums(dictionary: dict, dataclass_type):
    """
    This method converts literals to Enum in a ``dictionary`` of ``dataclass_type``
    Enum datatypes are determined from types annotation *without* forward reference.

    .. code-block:: python

        from tamm.utils import OptionalBool
        from tamm.utils.json._utils import JSONSerializableMixin

        @dataclasses.dataclass
        class MyDataclass(JSONSerializableMixin):

            # **[INCOMPATIBLE IMPLEMENTATION]**
            # Declaration style 1: will NOT deserialize to OptionalBool from
            # {"freeze_params": "TRUE"}
            freeze_params: "OptionalBool" = OptionalBool.NOTSET

             # Declaration style 2: WILL deserialize to OptionalBool from
            # {"freeze_params": "TRUE"}
            freeze_params: OptionalBool = OptionalBool.NOTSET

            @classmethod
            def _from_json_dict_impl(cls, **raw_dict: dict):
                raw_dict = maybe_get_enum_member_for_all_str_enums(raw_dict, cls)
                return super()._from_json_dict_impl(**raw_dict)

    Args:
        dictionary: raw dictionary for dataclass attribute
        dataclass_type: dataclass to initialize

    Returns: Dataclass instance of ``dataclass_type``

    """
    enum_fields = {
        f.name: f.type
        for f in dataclasses.fields(dataclass_type)
        if isinstance(f.type, enum_module.EnumMeta)
    }
    try:
        for enum_field, enum_class in enum_fields.items():
            dictionary[enum_field] = enum_class(dictionary[enum_field])
    except KeyError:
        # if dictionary[enum_field] is not found (i.e., some attribute of a
        # enum type not specified in the dictionary) this should still go through.
        pass
    except (ValueError, TypeError) as e:
        raise e
    return dictionary


def dataclass_init_drop_missing_keys(kwargs: Dict[str, Any], *, dataclass_type):
    """
    Returns the result of ``dataclass_type(**kwargs)`` but first filters any keys from
    kwargs that are not in the dataclass field names.

    .. warning::

        Use this *ONLY* on dataclasses like `ModelMetadata` which can leniently
        ignore newer keys.

    This method also converts literals to Enum for attributes of
    Enum types annotated *without* forward reference.

    Args:
        kwargs: raw dictionary for dataclass attribute
        dataclass_type: dataclass to initialize

    Returns: Dataclass instance of ``dataclass_type``

    """
    field_names = set(field.name for field in dataclasses.fields(dataclass_type))
    incompatible_keys = [key for key in kwargs if key not in field_names]
    kwargs = maybe_get_enum_member_for_all_str_enums(kwargs, dataclass_type)

    if len(incompatible_keys) > 0:
        kwargs = {key: val for key, val in kwargs.items() if key in field_names}
        logger.debug(
            "Ignoring incompatible keys %s when instantiating dataclass %s",
            incompatible_keys,
            dataclass_type,
        )
    return dataclass_type(**kwargs)


class DataClassReprMixin:
    def __repr__(self):
        obj = dataclass_to_dict(self)
        reprs = [f"{key}={repr(value)}," for key, value in obj.items()]
        joined = "\n".join(reprs)
        joined = textwrap.indent(joined, prefix=" " * 4)
        return f"{self.__class__.__name__}(\n{joined}\n)"


def clean_docstring(doc: Optional[str]) -> Union[str, None]:
    if doc is None:
        return None
    doc = inspect.cleandoc(doc)
    return doc.replace("\n", " ").replace("\r", " ")


def get_cleaned_docstring(func: Callable[..., Any]) -> Union[str, None]:
    doc = inspect.getdoc(func)
    return clean_docstring(doc)


def get_expanded_abspath(path):
    path = os.path.expanduser(path)
    path = os.path.expandvars(path)
    return os.path.abspath(path)


def size_in_bytes_to_string(num_bytes):
    # pylint: disable=invalid-name
    KiB = 1024
    MiB = KiB**2
    GiB = KiB**3
    if num_bytes > GiB:
        return f"{num_bytes / GiB:.1f} GiB"
    if num_bytes > MiB:
        return f"{num_bytes / MiB:.1f} MiB"
    if num_bytes > KiB:
        return f"{num_bytes / KiB:.1f} KiB"
    return f"{num_bytes} bytes"



from typing import Tuple as _Tuple

import torch as _torch
import torch.ao.quantization as _torchq
import torch.nn as _nn


class KVQuantizer(_nn.Module):
    """
    A layer for simulating the affect of KV cache quantization.

    Args:
        key_quantizer (:obj:`~.ao.layers.FakeQuantize`): A layer for quantizing
            attention keys.
        value_quantizer (:obj:`~.ao.layers.FakeQuantize`): A layer for quantizing
            attention values.
    """

    def __init__(
        self,
        *,
        key_quantizer: _torchq.FakeQuantizeBase,
        value_quantizer: _torchq.FakeQuantizeBase,
    ):
        super().__init__()
        self.key_quantizer = key_quantizer
        self.value_quantizer = value_quantizer

    def forward(
        self,
        query: _torch.Tensor,
        key: _torch.Tensor,
        value: _torch.Tensor,
    ) -> _Tuple[_torch.Tensor, _torch.Tensor, _torch.Tensor]:
        """
        Applies quantization to keys and values.

        Args:
            query: Queries for multi-headed attention (MHA).
            key: Keys for MHA.
            value: Values for MHA.

        Returns:
            A tuple ``(query, quantized_key, quantized_value)``, where ``quantized_key``
            and ``quantized_value`` are the result of calling ``key_quantizer`` and
            ``value_quantizer`` with the keys and values.
        """
        key = self.key_quantizer(key)
        value = self.value_quantizer(value)
        return query, key, value



import contextlib as _contextlib
from contextvars import ContextVar

first_token_generation = ContextVar("first_token_generation", default=False)


def get_first_token_generation_flag() -> bool:
    """
    Returns whether we are currently generating the first token in a sequence.

    Returns:
        bool: ``True`` if we're generating the first token, ``False`` otherwise
    """
    return first_token_generation.get()


@_contextlib.contextmanager
def first_token_generation_context():
    """
    Context manager for indicating first token generation.

    Sets a flag ``True`` indicating that we're generating the first token in a sequence.
    The flag is automatically reset to ``False`` when exiting the context.

    Example:
        >>> with first_token_generation_context():
        ...     first_token_logits = model(input_ids, ...)
    """
    token = first_token_generation.set(True)
    try:
        yield
    finally:
        first_token_generation.reset(token)



"""HF transformers model configuration for tamm SentencePiece tokenizer"""

import transformers as _transformers


class TammSentencePieceConfig(_transformers.PretrainedConfig):
    """
    A minimal config to go along with the :class:`.TammSentencePieceTokenizer`
    (since :mod:`transformers` requires a config to register the tokenizer
    with :class:`transformers.AutoTokenizer`).
    """

    model_type = "tamm_sentencepiece"



import importlib as _importlib
from typing import Optional as _Optional

import torch

from tamm.typing import StateDictType as _StateDictType


def _validate_missing_keys(state_dict: _StateDictType, layer_subset: _Optional[set]):
    missing_keys = layer_subset - set(state_dict.keys())
    if missing_keys:
        raise ValueError(
            f"layers '{missing_keys}' are missing in the state dict "
            f"but explicitly specified for checksum computation"
        )


def state_dict_checksum(
    state_dict: _StateDictType,
    layer_subset: _Optional[frozenset] = None,
    dtype=torch.float32,
    standardize_zero=False,
) -> str:
    """
    Computes 32-bit checksum of state dict using `xxhash <http://xxhash.com>`_,
    can optionally compute on a subset of layers

    .. note::

        Enabling standardize_zero may reduce the throughput by 70%

    Args:
        state_dict (:obj:`StateDictType`): Input state dictionary to verify checksum.
        layer_subset (:obj:`set`, optional): A set specifying selected keys (layers)
            in a state dict
        dtype (:obj:`torch.dtype`): torch dtype to compute checksum in
        standardize_zero (:obj:`bool`): normalize IEEE 754 +0/-0 to +0

    Returns:
         (:obj:`str`) Hash hex digest of ``state_dict``

    """
    try:
        xxhash = _importlib.import_module("xxhash")
    except ModuleNotFoundError as e:
        raise ModuleNotFoundError("'xxhash' is required for model checksum.") from e

    hash_alg = xxhash.xxh3_64()
    if layer_subset is not None:
        _validate_missing_keys(state_dict, layer_subset)
        state_dict = {
            layer: v for layer, v in state_dict.items() if layer in layer_subset
        }

    state_dict = dict(sorted(state_dict.items()))

    for v in state_dict.values():
        v = v.to(dtype=dtype)
        if standardize_zero:
            v[v == 0] = 0
        hash_alg.update(v.numpy(force=True).tobytes(order="C"))
    return hash_alg.hexdigest()



from torch import nn as _nn


def get_num_params(module: _nn.Module, *, recurse: bool = True) -> int:
    """
    Returns the total number of parameter elements in a :obj:`torch.nn.Module`.

    Args:
        module (:obj:`nn.Module`): The model or layer.
        recurse (:obj:`bool`, optional): A flag that controls recursion into
            submodules.  When ``False``, the function counts only parameters
            that are direct members of ``module``.  When ``True``, the function
            counts parameters from ``module`` and its submodules.  Defaults to
            ``True``.
    """
    return sum(p.numel() for p in module.parameters(recurse=recurse))


def get_num_trainable_params(module: _nn.Module, *, recurse: bool = True) -> int:
    """
    Returns the number of parameters in a module for which
    ``requires_grad`` is ``True``.

    Args:
        module (:obj:`nn.Module`): The model or layer.
        recurse (:obj:`bool`, optional): A flag that controls recursion into
            submodules.  Defaults to ``True``.
    """
    return sum(p.numel() for p in module.parameters(recurse=recurse) if p.requires_grad)


def get_num_frozen_params(module: _nn.Module, *, recurse: bool = True) -> int:
    """
    Returns the number of parameters in a module for which
    ``requires_grad`` is ``False``.

    Args:
        module (:obj:`nn.Module`): The model or layer.
        recurse (:obj:`bool`, optional): A flag that controls recursion into
            submodules.  Defaults to ``True``.
    """
    return sum(
        p.numel() for p in module.parameters(recurse=recurse) if not p.requires_grad
    )



from tamm.layers.mixture_of_experts.dispatcher.capacity_constrained import (
    CapacityConstrainedMixtureOfExpertsDispatcher,
)
from tamm.layers.mixture_of_experts.dispatcher.capacity_constraint import (
    MixtureOfExpertsCapacityConstraint,
    MixtureOfExpertsCapacityConstraintOutput,
    VanillaMixtureOfExpertsCapacityConstraint,
)
from tamm.layers.mixture_of_experts.dispatcher.common import (
    MixtureOfExpertsDispatcher,
    MixtureOfExpertsDispatcherForwardMode,
    MixtureOfExpertsDispatchOutput,
)
from tamm.layers.mixture_of_experts.dispatcher.dropless import (
    DroplessMixtureOfExpertsDispatcher,
)



"""
tamm.ao
-------

This module implements building blocks for architecture optimizations, similar to
:mod:`torch.ao`.  Activation quantization is one type of supported optimization.

.. automodule:: tamm.ao.arch_optimizers

.. automodule:: tamm.ao.layers

.. automodule:: tamm.ao.utils
"""

from tamm.ao import arch_optimizers, layers, utils
from tamm.ao.arch_optimizers import ArchOptimizer, KVQuantArchOptimizer

__all__ = [
    "ArchOptimizer",
    "KVQuantArchOptimizer",
    "arch_optimizers",
    "layers",
    "utils",
]



"""
utils.torch_utils
=================

This module implements utilities for working with PyTorch.


Iteration tools
^^^^^^^^^^^^^^^

.. autofunction:: tamm.utils.torch_utils.iter_named_parameters_and_buffers_with_layers

.. autofunction:: tamm.utils.torch_utils.iter_named_parameters_and_buffers

.. autofunction:: tamm.utils.torch_utils.map_named_parameters_and_buffers


Parameter counting
^^^^^^^^^^^^^^^^^^

.. autofunction:: tamm.utils.torch_utils.get_num_params

.. autofunction:: tamm.utils.torch_utils.get_num_frozen_params

.. autofunction:: tamm.utils.torch_utils.get_num_trainable_params


Exportable dataclass
^^^^^^^^^^^^^^^^^^^^

.. autofunction:: tamm.utils.torch_utils.torch_exportable_dataclass


Parameter tying
^^^^^^^^^^^^^^^

.. autoclass:: tamm.utils.torch_utils.TieableParamsDict
    :members: register_tied_parameter, get_tied_value
"""

from tamm.utils.torch_utils.exportable_dataclass import torch_exportable_dataclass
from tamm.utils.torch_utils.itertools import (
    iter_named_parameters_and_buffers,
    iter_named_parameters_and_buffers_with_layers,
    map_named_parameters_and_buffers,
)
from tamm.utils.torch_utils.num_params import (
    get_num_frozen_params,
    get_num_params,
    get_num_trainable_params,
)
from tamm.utils.torch_utils.tieable_params import TieableParamsDict



import copy
import logging as _logging

from tamm.model_repo._warning import _warn_deprecated_published_config

logger = _logging.getLogger(__name__)


def get_model_config_from_any_tamm_object(any_tamm_object):
    model_config = any_tamm_object
    if any_tamm_object.__class__.__name__ == "PublishedModelConfig":
        _warn_deprecated_published_config(any_tamm_object)
        model_config = any_tamm_object.model_config
    # pylint: disable=import-outside-toplevel
    from tamm.models.common import ModelConfig

    if not isinstance(model_config, ModelConfig):
        raise ValueError(
            f"Unsupported model config type: {type(model_config).__name__}"
        )
    return copy.deepcopy(model_config)



from typing import Callable

from torch import nn as _nn


class CompositePostHook:
    def __init__(self, *hooks: "Callable"):
        self.hooks = hooks

    def __call__(self, model: "_nn.Module") -> "_nn.Module":
        for hook in self.hooks:
            model = hook(model)
        return model


class IdentityPostHook:
    def __call__(self, x):
        return x



import abc as _abc
from typing import List as _List
from typing import Protocol as _Protocol
from typing import Tuple as _Tuple
from typing import Union as _Union

import torch as _torch


class BaseKVCache(_Protocol):
    """
    A :obj:`Protocol` that defines an abstract interface between the inference runner
    and its key-value cache.
    """

    @property
    @_abc.abstractmethod
    def length(self) -> int:
        """The sequence length that the cache is initialized for."""

    @_abc.abstractmethod
    def up_to_index(self, index: int) -> "BaseKVCacheView":
        """
        Returns a :obj:`.BaseKVCacheView` that exposes the first ``index`` tokens
        of the cache.  This enables pre-allocation of the cache with a large sequence
        length and then efficient incremental writes while filling the cache.

        Args:
            index (:obj:`int`): The length of the resulting cache view.  This must not
                exceed the length of the cache.

        Returns:
            A cache view of the first ``index`` tokens in the cache.
        """

    @_abc.abstractmethod
    def resize(self, new_length: int):
        """
        Updates the underlying sequence length of the cache while keeping previously
        cached values.

        Args:
            new_length (:obj:`int`): The new sequence length.
        """

    @_abc.abstractmethod
    def to(self, *args, **kwargs) -> "BaseKVCache":  # pylint: disable=invalid-name
        """
        Change the device and/or dtype of the cache, similar to :func:`torch.Tensor.to`.

        Returns:
            The :obj:`BaseKVCache` instance (``self``).
        """


class BaseKVCacheView(_Protocol):
    """
    A :obj:`Protocol` that defines an abstract interface between a
    :obj:`.TransformerStack` and its key-value cache.
    """

    @property
    @_abc.abstractmethod
    def length(self) -> int:
        """The sequence length of the cache view."""

    @property
    @_abc.abstractmethod
    def num_layers(self) -> int:
        """The number of layers in the cache view."""

    @property
    @_abc.abstractmethod
    def dtype(self) -> _Union[_torch.dtype, None]:
        """
        The :obj:`torch.dtype` of the cached values.  May return ``None`` if the
        cache is still empty.
        """

    @property
    @_abc.abstractmethod
    def device(self) -> _torch.device:
        """The :obj:`torch.device` of the cached values."""

    def at_layer(self, layer_idx: int) -> "KVCacheLayerView":
        """
        Exposes the cache for a specific layer of the model.

        Args:
            layer_idx (:obj:`int`): The layer index.  This must be less than
                ``num_layers``.

        Returns:
            A :obj:`KVCacheLayerView` for layer ``layer_idx`` of the model/cache.
        """
        return KVCacheLayerView(kv_cache_view=self, layer_idx=layer_idx)

    def at_layers(self, layer_indices: _List[int]) -> "KVCacheLayersView":
        return KVCacheLayersView(kv_cache_view=self, layer_indices=layer_indices)

    @_abc.abstractmethod
    def write_to_tail(
        self, key: _torch.Tensor, value: _torch.Tensor, *, layer_idx: int
    ) -> None:
        """
        Writes new key and value tensors to the cache for a specific layer, overwriting
        the trailing ``num_new_tokens`` cached keys and values for this layer.

        Args:
            key (:obj:`torch.Tensor`): A newly computed key tensor with shape
                ``(batch_size, num_new_tokens, num_kv_heads * dim_per_head)``.
            value (:obj:`torch.Tensor`): A newly computed value tensor with shape
                ``(batch_size, num_new_tokens, num_kv_heads * dim_per_head)``.
            layer_idx (:obj:`int`): The index of the layer in the model/cache.
        """

    def write_to_tail_vectorized(
        self, keys: _torch.Tensor, values: _torch.Tensor, *, layer_indices: _List[int]
    ) -> None:
        # Note: It's up to the specific KVCacheView to provide efficient vectorized implementation
        # Note that dimension after batch (0th indx) is the vectorize dimension
        for idx, layer_idx in enumerate(layer_indices):
            self.write_to_tail(
                _torch.select(keys, 1, idx),
                _torch.select(values, 1, idx),
                layer_idx=layer_idx,
            )

    @_abc.abstractmethod
    def write_segment_ids_to_tail(self, segment_ids: _torch.Tensor) -> None:
        """
        Writes new segment IDs to the cache, overwriting the trailing ``num_new_tokens``
        cached segment IDs.

        Args:
            segment_ids (:obj:`torch.Tensor`): An integer tensor of segment IDs with
                shape ``(batch_size, num_new_tokens)``.
        """

    @_abc.abstractmethod
    def write_positions_to_tail(self, positions: _torch.Tensor) -> None:
        """
        Writes new positions to the cache, overwriting the trailing ``num_new_tokens``
        cached positions.

        Args:
            positions (:obj:`torch.Tensor`): An integer tensor of positions with
                shape ``(batch_size, num_new_tokens)``.
        """

    def write_token_types_to_tail(
        self, token_types: _torch.Tensor
    ) -> None:  # no @abstractmethod since most models don't call it
        """
        Writes new token types to the cache, overwriting the trailing ``num_new_tokens``
        cached positions.

        Args:
            positions (:obj:`torch.Tensor`): An integer tensor of positions with
                shape ``(batch_size, num_new_tokens)``.
        """
        raise NotImplementedError(
            f"{self.__class__} has not implemented write_token_types_to_tail()"
        )

    @_abc.abstractmethod
    def read(self, layer_idx: int) -> _Tuple[_torch.Tensor, _torch.Tensor]:
        """
        Reads keys and values from the cache for a specific layer.

        Args:
            layer_idx (:obj:`int`): The index of the layer in the model/cache.

        Returns:
            A :obj:`tuple` of two tensors, each with shape
            ``(batch_size, length, num_kv_heads * dim_per_head)``.  The first
            element is the keys and the second is the values.  The ``length`` is the
            full sequence length of the cache view.
        """

    def read_batched(
        self, layer_indices: _List[int]
    ) -> _Tuple[_torch.Tensor, _torch.Tensor]:
        """
        Reads keys and values from the caches.

        Returns:
            A :obj:`tuple` of lists, each tensor within the list has shape
            ``(batch_size, length, num_kv_heads * dim_per_head)``.  The first
            set of elements are the keys and the second are the values.  The ``length`` is the
            full sequence length of the cache view.
        """
        qs, vs = [], []
        for layer_idx in layer_indices:
            q, v = self.read(layer_idx)
            qs.append(q)
            vs.append(v)
        return _torch.stack(qs, dim=1), _torch.stack(vs, dim=1)

    @_abc.abstractmethod
    def read_segment_ids(self) -> _torch.Tensor:
        """
        Reads segment IDs from the cache.

        Returns:
            An integer tensor of segment IDs with shape ``(batch_size, length)``, where
            ``length`` is the full sequence length of the cache view.
        """

    @_abc.abstractmethod
    def read_positions(self) -> _torch.Tensor:
        """
        Reads positions from the cache.

        Returns:
            An integer tensor of positions with shape ``(batch_size, length)``, where
            ``length`` is the full sequence length of the cache view.
        """

    def read_token_types(
        self,
    ) -> _torch.Tensor:  # no @abstractmethod since most models don't call it
        """
        Reads token types from the cache.

        Returns:
            An integer tensor of token types with shape ``(batch_size, length)``, where
            ``length`` is the full sequence length of the cache view.
        """
        raise NotImplementedError(
            f"{self.__class__} has not implemented read_token_types()"
        )


class KVCacheLayerView:
    """
    A class for reading and writing keys and values for a specific layer of a
    :class:`.BaseKVCacheView`.
    """

    def __init__(self, *, kv_cache_view: BaseKVCacheView, layer_idx: int) -> None:
        self.kv_cache_view = kv_cache_view
        self.layer_idx = layer_idx

    @property
    def dtype(self) -> _torch.dtype:
        """The :obj:`torch.dtype` of the cached values."""
        return self.kv_cache_view.dtype

    @property
    def device(self) -> _torch.device:
        """The :obj:`torch.device` of the cached values."""
        return self.kv_cache_view.device

    def write_to_tail(self, key: _torch.Tensor, value: _torch.Tensor) -> None:
        """
        Writes new key and value tensors to the cache, overwriting the trailing
        ``num_new_tokens`` cached keys and values for this layer.

        Args:
            key (:obj:`torch.Tensor`): A newly computed key tensor with shape
                ``(batch_size, num_new_tokens, num_kv_heads * dim_per_head)``.
            value (:obj:`torch.Tensor`): A newly computed value tensor with shape
                ``(batch_size, num_new_tokens, num_kv_heads * dim_per_head)``.
        """
        if self.dtype is not None:
            are_dtypes_matching = self.dtype == key.dtype == value.dtype
            if not are_dtypes_matching:
                raise TypeError(
                    "dtype mismatch in kv cache "
                    f"(key: {key.dtype}, value: {value.dtype}, cache: {self.dtype})"
                )

        self.kv_cache_view.write_to_tail(key, value, layer_idx=self.layer_idx)

    def read(self) -> _Tuple[_torch.Tensor, _torch.Tensor]:
        """
        Reads keys and values from the cache.

        Returns:
            A :obj:`tuple` of two tensors, each with shape
            ``(batch_size, length, num_kv_heads * dim_per_head)``.  The first
            element is the keys and the second is the values.  The ``length`` is the
            full sequence length of the cache view.
        """
        return self.kv_cache_view.read(self.layer_idx)


class KVCacheLayersView:
    """
    A class for reading and writing keys and values for a _multiple_ layers of a
    :class:`.BaseKVCacheView`.
    """

    def __init__(
        self, *, kv_cache_view: BaseKVCacheView, layer_indices: _List[int]
    ) -> None:
        self.kv_cache_view = kv_cache_view
        self.layer_indices = layer_indices

    @property
    def dtype(self) -> _torch.dtype:
        """The :obj:`torch.dtype` of the cached values."""
        return self.kv_cache_view.dtype

    @property
    def device(self) -> _torch.device:
        """The :obj:`torch.device` of the cached values."""
        return self.kv_cache_view.device

    def write_to_tail(self, keys: _torch.Tensor, values: _torch.Tensor) -> None:
        """
        Writes new key and value tensors to the caches, overwriting the trailing
        ``num_new_tokens`` cached keys and values for this layer.

        Args:
            keys (:obj: `torch.Tensor`): A newly computed stacked list of key tensors with shape
                ``(batch_size, vec_dim, num_new_tokens, num_kv_heads * dim_per_head)``.
            values (:obj: `torch.Tensor`): A newly computed stacked list of value tensor with shape
                ``(batch_size, vec_dim, num_new_tokens, num_kv_heads * dim_per_head)``.
        """
        if self.dtype is not None:
            are_dtypes_matching = self.dtype == keys.dtype == values.dtype
            if not are_dtypes_matching:
                raise TypeError(
                    "dtype mismatch in kv cache "
                    f"(keys: {keys.dtype}, values: {values.dtype}, cache: {self.dtype})"
                )

        self.kv_cache_view.write_to_tail_vectorized(
            keys, values, layer_indices=self.layer_indices
        )

    def read(self) -> _Tuple[_torch.Tensor, _torch.Tensor]:
        """
        Reads keys and values from the caches.

        Returns:
            A :obj:`tuple` of lists, each tensor within the list has shape
            ``(batch_size, length, num_kv_heads * dim_per_head)``.  The first
            set of elements are the keys and the second are the values.  The ``length`` is the
            full sequence length of the cache view.
        """
        return self.kv_cache_view.read_batched(layer_indices=self.layer_indices)



"""
Common functionality for the cli, including command and argument decorators for
defining new commands.
"""

import argparse
import functools
from typing import Optional

from tamm import _plugin
from tamm.context_vars import _plugin_init

PARSERS = {}
HEADER = r"""
  __
_/  |_ _____     _____    _____
\   __\\__  \   /     \  /     \
 |  |   / __ \_|  | |  \|  | |  \
 |__|  (______/|__|_|__/|__|_|__/

"""


def command(*command_path, help_text: Optional[str] = None):
    """
    A decorator for registering a command with the cli.

    Args:
        *command_path: strings that specify the command path.
        help_text: a brief description of the command.

    Example for adding a ``> tamm print hello`` subcommand:

        @command("print", "hello", help_text="Prints hello")
        def print_hello():
            print("Hello!")
    """

    def decorator(fn):
        if not isinstance(fn, Command):
            fn = Command(fn)
        fn.register(*command_path, help_text=help_text)
        return fn

    return decorator


def argument(*args, **kwargs):
    """
    A decorator for adding arguments to the cli command.

    Args:
        *args, **kwargs: Arguments for ArgumentParser.add_argument()
    """

    def decorator(fn):
        if not isinstance(fn, Command):
            fn = Command(fn)
        fn.add_argument(*args, **kwargs)
        return fn

    return decorator


def register_namespace_help_text(*command_path, help_text: str) -> None:
    """
    Register help text for a cli command path that does not implement
    a command function.

    Args:
        *command_path: strings that specify the command path.
        help_text: a brief description of the command.

    Example for a "publish" namespace:

        register_namespace_help_text(
            "publish", help_text="Commands for publishing objects"
        )

        @command("publish", "model", help_text="Publish a model")
        def publish_model():
            ...

        @command("publish", "preprocessor", help_text="Publish a preprocessor")
        def publish_model():
            ...
    """
    get_parser(*command_path, help_text=help_text)


def get_parser(*command_path, help_text=None):
    """
    Returns the parser for a subcommand.

    Args:
        *command_path: Strings that specify the command path.
    """
    try:
        return PARSERS[command_path]
    except KeyError:
        if not command_path:
            parser = argparse.ArgumentParser(prog="tamm")
        else:
            if len(command_path) == 1:
                help_text = _maybe_update_help_text_with_plugin_name(help_text)
            parent_parser = get_parser(*command_path[:-1])
            subparser = get_subparsers(parent_parser)
            parser = subparser.add_parser(command_path[-1], help=help_text)
        PARSERS[command_path] = parser
        return PARSERS[command_path]


def get_extended_root_parser() -> argparse.ArgumentParser:
    """Returns the root parser extended with plugin commands."""
    _plugin.register_cli_commands()
    return get_parser()


def _maybe_update_help_text_with_plugin_name(help_text: str):
    plugin_name = _plugin_init.get_initializing_plugin_module_name()
    if plugin_name is None:
        return help_text
    return f"{help_text} ({plugin_name} plugin command)"


@functools.cache
def get_subparsers(parser):
    """Creates and returns the subparsers object for a parser."""
    return parser.add_subparsers(title="subcommands")


class Command:
    """A class for wrapping a function and turning it into a tamm cli command."""

    def __init__(self, fn):
        self.fn = fn
        functools.update_wrapper(self, fn)
        self.arguments = []

    def register(self, *command_path, help_text=None):
        """Registers the command with the CLI"""
        parser = get_parser(*command_path, help_text=help_text)
        parser.set_defaults(func=self)
        for args, kwargs in self.arguments:
            parser.add_argument(*args, **kwargs)
        self.arguments = None

    def add_argument(self, *add_argument_args, **add_argument_kwargs):
        """
        Adds arguments to the command.

        Args:
            *args, **kwargs: Arguments for ArgumentParser.add_argument()
        """
        self.arguments.append((add_argument_args, add_argument_kwargs))

    def __call__(self, *args, **kwargs):
        return self.fn(*args, **kwargs)



import abc as _abc
import contextlib as _contextlib
import inspect as _inspect
import logging as _logging
from typing import TYPE_CHECKING
from typing import Dict as _Dict
from typing import Optional as _Optional
from typing import Union as _Union

from tamm import _helpers
from tamm.layers.common.builder import BuildableMixin as _BuildableMixin
from tamm.layers.common.config import LayerConfig
from tamm.layers.common.context_hooks import (
    DefaultDeviceContextHook,
    DtypeContextHook,
    FreezeParamsContextHook,
    PretrainedContextHook,
    UseMetaInitTrickContextHook,
)
from tamm.layers.common.post_hooks import (
    ArchOptimizersPostHook,
    FreezeParamsPostHook,
    IdentityPostHook,
    get_model_adapters_post_hook,
)
from tamm.layers.common.post_hooks.model_initializer import ModelInitializerPostHook
from tamm.utils import OptionalBool as _OptionalBool

if TYPE_CHECKING:
    from tamm._adapters_v1 import ModelAdapter
    from tamm.ao import ArchOptimizer
    from tamm.typing import (
        LenientOptionalBool,
        OptionalDeviceOrString,
        OptionalDtypeOrString,
    )
_logger = _logging.getLogger(__name__)


class LayerMixin(_BuildableMixin):
    """
    Common layer mixin for all |tamm| layers.
    All layers are expected to inherit from this class

    This mixin makes layer

    1. Buildable: layer inherits this class has attribute ``.Builder``.
    """


class _BaseConfigurableMixin(LayerMixin, _abc.ABC):
    def __init_subclass__(cls, /, **kwargs):
        super().__init_subclass__(**kwargs)
        if not _inspect.ismethod(cls.create_basic_builder):
            raise TypeError(
                f"Expected {cls.__name__}.create_basic_builder() to be a classmethod"
            )

    @classmethod
    @_abc.abstractmethod
    def _should_update_signature(cls):
        """
        This predicate determines whether the signature for create_builder() should
        be updated
        """

    @classmethod
    def _update_create_signature(cls):
        new_func = _helpers.passthrough_decorator(cls.create.__func__)
        new_func.__signature__ = _inspect.signature(cls.create_builder.__func__)
        cls.create = classmethod(new_func)

    @classmethod
    def _update_create_builder_signature(cls):
        """
        This classmethod applies a wrapper to the create_builder() classmethod to
        update its signature.  This is important because we derive the fields for
        each module's Config from the signature of create_builder().

        create_builder() takes its own keyword arguments but it also accepts *args and
        **kwargs for create_basic_builder().  To create the signature for
        create_builder(), we take the signature of create_basic_builder() and append
        the keyword arguments from create_builder().
        """

        basic_sig = _inspect.signature(cls.create_basic_builder.__func__)
        sig = cls._get_base_create_builder_signature()
        common_params = [p for p in sig.parameters.values() if p.kind is p.KEYWORD_ONLY]

        for param in common_params:
            if param.name in basic_sig.parameters:
                raise RuntimeError(
                    f"{cls.__name__}.create_basic_builder() contains a parameter named "
                    f"'{param.name}', but this is reserved for create_builder(), which "
                    "wraps create_basic_builder()."
                )

        new_sig = _helpers.add_kw_only_params_to_signature(basic_sig, common_params)
        new_func = _helpers.passthrough_decorator(cls.create_builder.__func__)
        new_func.__signature__ = new_sig
        cls.create_builder = classmethod(new_func)

    @staticmethod
    def _get_base_create_builder_signature():
        return _inspect.signature(_BaseConfigurableMixin.create_builder.__func__)

    @classmethod
    @_abc.abstractmethod
    def create_basic_builder(cls):
        """
        Creates and returns a configured instance of the module's
        :class:`~.layers.LayerBuilder` subclass. subclasses should
        implement this method and add keyword arguments for configuring module options
        (such as the number of layers, dtype of parameters, etc.).  These keyword
        arguments cannot include keyword arguments from ``.create_builder``
        (``device``, ``pretrained_path``, etc.), since ``.create_builder``
        wraps ``.create_basic_builder()`` and adds additional functionality.
        """

    @classmethod
    def create_builder(
        cls,
        *basic_builder_args,
        device: "OptionalDeviceOrString" = None,
        adapters: _Optional[_Dict[str, "ModelAdapter"]] = None,
        active_adapter: _Union[int, str] = 0,
        arch_optimizers: _Optional[_Dict[str, "ArchOptimizer"]] = None,
        pretrained: "LenientOptionalBool" = _OptionalBool.NOTSET,
        pretrained_path: _Optional[str] = None,
        dtype: "OptionalDtypeOrString" = None,
        freeze_params: "LenientOptionalBool" = _OptionalBool.NOTSET,
        **basic_builder_kwargs,
    ):
        # pylint: disable=too-many-locals
        """
        Creates and returns a configured instance of the module's
        :class:`~.layers.LayerBuilder` subclass.  This method adds hooks to the result
        of ``.create_basic_builder()`` to add common functionality related to module
        instantiation.

        Args:
            device (:obj:`torch.device` or :obj:`str`, optional): The target device for
                module parameters and other state.  This is also the default device for
                adapter layers, which can be overridden in :obj:`.ModelAdapter`.
            adapters (:obj:`dict`, optional): A dictionary that maps :obj:`str` adapter
                IDs to instances of :obj:`.ModelAdapter`.  The returned builder will
                call these adapters to attach adapter layers (such as LoRA) to newly
                created modules.
            active_adapter (:obj:`int` or `str`, optional): A variable indicates which
                adapter should be activated when the model is being instantiated.
                If the value is an integer, that stands for the index of the adapter.
                If the value is a string, it should be a valid adapter id.
            arch_optimizers (:obj:`dict`, optional): A dictionary that maps :obj:`str`
                architecture optimizer IDs to instances of
                :obj:`.ArchOptimizer`. The returned builder will apply these
                architecture optimizers (such as KVQuant) to newly created modules.
            pretrained (:obj:`OptionalBool`, `bool`): A flag for loading pre-trained
                weights.  The module only loads from pretrained paths when
                ``pretrained`` is ``True``. This same flag also applies to the module
                adapters when initializing adapters.
            pretrained_path (:obj:`str`, optional): A pretrained checkpoint path for the
                backbone module's state dict.  This state should not include adapter
                state.
            freeze_params (:obj:`OptionalBool`, `bool`): A flag for freezing parameters
                of the module (i.e., setting their ``requires_grad`` attributes to
                ``False``).  Note: this flag only applies to the backbone module,
                not adapter parameters, since adapters have their own
                ``freeze_params`` option.
            dtype (:obj:`torch.dtype`, optional): The parameter dtype.

        Returns:
            The configured :obj:`.LayerBuilder`.
        """
        # Type casting first since ``pretrained`` can be either `bool` or `OptionalBool`
        disable_meta_device_init_trick = (
            _OptionalBool(pretrained) is _OptionalBool.FALSE
        )
        extended_context_hooks = [
            DefaultDeviceContextHook(
                device=device,
                disable_meta_device_init_trick=disable_meta_device_init_trick,
            ),
            PretrainedContextHook(pretrained),
            DtypeContextHook(dtype),
            FreezeParamsContextHook(freeze_params),
        ]
        with _contextlib.ExitStack() as stack:
            for extended_context_hook in extended_context_hooks:
                # Apply the same set of extended contexts to the calling context of
                # `create_basic_builder`, in case module developers instantiate concrete
                # ``nn.Modules`` within the builder.
                # fixme: Debate whether to un-support this use case and enforce builder
                # to only compose other 'pure' builders
                # (i.e., no nn.Module gets created 'within' a builder)
                stack.enter_context(extended_context_hook())

            builder = cls.create_basic_builder(
                *basic_builder_args, **basic_builder_kwargs
            )

        # UseMetaInitTrickContextHook is the only 'narrow scoped' context hook
        # (covering builder() only)
        builder.context_hooks.register(
            UseMetaInitTrickContextHook(
                pretrained=pretrained, pretrained_path=pretrained_path
            )
        )

        for extended_context_hook in extended_context_hooks:
            # Register 'extended' context-hooks (covering both builder() and post_hooks)
            builder.extended_context_hooks.register(extended_context_hook)

        post_hooks = [
            ModelInitializerPostHook(pretrained_path),
            FreezeParamsPostHook(),
            get_model_adapters_post_hook(
                adapters=adapters,
                active_adapter=active_adapter,
            ),
            ArchOptimizersPostHook(
                arch_optimizers=arch_optimizers,
            ),
        ]
        for post_hook in post_hooks:
            # fixme: refactor dataclass partial's hook registration API to filter out
            #  IdentityPostHook
            if not isinstance(post_hook, IdentityPostHook):
                builder.post_hooks.register(post_hook)

        return builder

    @classmethod
    def create(cls, *args, **kwargs):
        """
        Creates and returns a configured instance of the module. This method accepts the
        same arguments as :meth:`.create_builder`.
        """
        builder = cls.create_builder(*args, **kwargs)
        return builder.build()


class ConfigurableLayerMixin(_BaseConfigurableMixin):
    """
    Common layer mixin for all configurable |tamm| layers.

    This mixin makes layer

    1. Buildable: layer inherits this class has attribute ``.Builder``.
    2. Configurable: <layer_cls>.Config() instantiates a Config object representing the
       layer itself. ``<layer_cls>.Config().create_layer(*args, **kwargs)`` creates an
       instance of the layer with optional argument overrides ``*args`` and ``**kwargs``

    """

    def __init_subclass__(cls, /, **kwargs):
        super().__init_subclass__(**kwargs)
        if cls._should_update_signature():
            cls._update_create_builder_signature()
            cls._update_create_signature()
        cls.Config = LayerConfig._create_subclass_from_module_cls(cls)

    @classmethod
    def _should_update_signature(cls):
        if "create_builder" in cls.__dict__:
            _logger.debug(
                f"Skipping update to the signature of {cls.__name__}.create_builder() "
                "because this class overrides create_builder().  This may impact the "
                f"LayerConfig type for {cls}."
            )
            return False
        return True

    @classmethod
    @_abc.abstractmethod
    def create_basic_builder(cls):
        """
        Creates and returns a configured instance of the layer's
        :class:`~.layers.LayerBuilder` subclass. :class:`.ConfigurableLayerMixin`
        subclasses should implement this method and add keyword arguments for
        configuring layer options (such as the number of layers, dtype of parameters,
        etc.).  These keyword arguments cannot include keyword arguments from
        :meth:`_BaseConfigurableMixin.create_builder` (``device``, ``pretrained_path``,
        etc.), since :meth:`_BaseConfigurableMixin.create_builder` wraps
        ``.create_basic_builder()`` and adds additional functionality.
        """



import contextlib as _contextlib
from contextvars import ContextVar
from typing import Union

from tamm.utils.optional_bool import OptionalBool

freeze_params = ContextVar("freeze_params", default=OptionalBool.NOTSET)


def get_default_freeze_params_flag():
    """
    ``.create_builder()`` of each configurable module will use the return value to
    resolve whether to freeze parameters, *if* invoked without an explict
    Boolean value (i.e., ``None``, or ``OptionalBool.NOTSET``).

    Default value is ``False``
    """
    return freeze_params.get()


def set_default_freeze_params_flag(value: Union[bool, OptionalBool]):
    """
    Sets |tamm| global ``freeze_params`` value

    Args:
        value: True, False. None, OptionalBool.[TRUE, FALSE, NOTSET]
    """
    value = OptionalBool(value)
    return freeze_params.set(value)


@_contextlib.contextmanager
def freeze_params_flag_context(value):
    """
    Context of |tamm| global ``freeze_params``

    Example:

    .. code-block:: python

        with freeze_params_context(True):
            tamm.create_model(...) # freeze_params=True not needed

    """
    token = set_default_freeze_params_flag(value)
    try:
        yield
    finally:
        freeze_params.reset(token)



from enum import Enum as _Enum


class OptionalBool(str, _Enum):
    TRUE = "TRUE"
    FALSE = "FALSE"
    NOTSET = "NOTSET"

    @classmethod
    def _missing_(cls, value):
        if isinstance(value, str):
            value = value.upper()
            for member in cls:
                if member.value == value:
                    return member
        if value is True:
            return OptionalBool.TRUE
        if value is False:
            return OptionalBool.FALSE
        if value is None:
            return OptionalBool.NOTSET
        return None

    def __bool__(self):
        if self is OptionalBool.TRUE:
            return True
        if self is OptionalBool.FALSE:
            return False

        raise ValueError("OptionalBool.notset does not have a truth value")

    def __eq__(self, other) -> bool:
        """Overrides the default implementation"""
        if isinstance(other, OptionalBool):
            return self is other
        if isinstance(other, bool):
            return False if self is self.NOTSET else bool(self) is other
        if other is None:
            return self is OptionalBool.NOTSET

        raise ValueError("OptionalBool.notset can only compare with bool or None")

    def __ne__(self, other) -> bool:
        return not self == other

    def __repr__(self):
        return self.value

    def __hash__(self):
        return hash(self.value)

    def __invert__(self):
        if self is OptionalBool.TRUE:
            return OptionalBool.FALSE
        if self is OptionalBool.FALSE:
            return OptionalBool.TRUE
        raise ValueError(f"Cannot invert {self}")



from tamm._adapters_v1 import layer_adapters
from tamm._adapters_v1.adapted_layer import AdaptedLayer
from tamm._adapters_v1.adapter_api import (
    AdaptedModelStateDictFormat,
    delete_adapter,
    freeze_adapter,
    get_all_active_adapter_ids,
    get_all_adapter_ids,
    has_adapter,
    init,
    is_adapter_initialized,
    load_state_dict,
    map_state_dict,
    merge_adapter,
    set_active_adapter,
    unfreeze_adapter,
    unset_active_adapter,
)
from tamm._adapters_v1.layer_adapters import LayerAdapter, LayerAdapterConfig
from tamm._adapters_v1.layer_annotations import annotate_layer
from tamm._adapters_v1.model_adapters.composition import (
    AveragedInputTransformsModelAdapter,
    AveragedOutputTransformsModelAdapter,
    StackedInputTransformsModelAdapter,
    StackedOutputTransformsModelAdapter,
)
from tamm._adapters_v1.model_adapters.lora import (
    LoRAModelAdapter,
    MultiLoRAModelAdapter,
    SoftMixingLoRAModelAdapter,
)
from tamm._adapters_v1.model_adapters.model_adapter import ModelAdapter
from tamm._adapters_v1.utils import (
    create_v1_state_dict,
    find_adapter_ids_from_state_dict,
    get_linear_adapter,
    get_num_adapter_params,
    merge_adapters,
    split_adapted_weights,
)



"""
generation.token_generators
===========================

.. autoclass:: tamm.generation.token_generators.TokenGenerator
    :members:
    :inherited-members:

.. autoclass:: tamm.generation.token_generators.GreedyTokenGenerator
    :members:
    :inherited-members:

.. autoclass:: tamm.generation.token_generators.SamplingTokenGenerator
    :members:
    :inherited-members:
"""

from tamm.generation.token_generators.base import TokenGenerator
from tamm.generation.token_generators.greedy import GreedyTokenGenerator
from tamm.generation.token_generators.sampling import SamplingTokenGenerator
from tamm.generation.token_generators.speculative import (
    SpeculativeDecodingTokenGenerator,
)



import abc
import copy
import json
import logging
from itertools import chain
from typing import Any, List, Sequence, Union

import tamm.utils.json as tamm_json
from tamm.preprocessors.base import Preprocessor

logger = logging.getLogger(__name__)


class PreprocessorRegistry(abc.ABC):
    # pylint: disable-next=unused-argument
    def __init__(self, *args, **kwargs):
        del args, kwargs

    def create(self, identifier: Any, *args, **kwargs) -> "Preprocessor":
        """
        Creates a preprocessor referred by ``identifier``
        Args:
            identifier: :class:`Preprocessor` or the name of the preprocessor
            **kwargs: Extra keyword arguments
        Returns: Preprocessor instance
        """
        if isinstance(identifier, Preprocessor):
            preprocessor = identifier
        else:
            preprocessor = self._create(identifier, *args, **kwargs)
        # always use a shallow copy so that instances do not share configurable attributes
        preprocessor = copy.copy(preprocessor)
        for k, v in kwargs.items():
            setattr(preprocessor, k, v)
        return preprocessor

    def clear_cache(self):
        """
        Clears any local cache if setup by the concrete catalog
        """

    @abc.abstractmethod
    def _create(self, identifier: Any, *args, **kwargs) -> "Preprocessor":
        """
        Abstract method for creating a preprocessor.
        Shall be implemented by concrete class.
        """

    @abc.abstractmethod
    def list_objects(self, *args, **kwargs) -> List[str]:
        """
        Returns a list of preprocessors available
        *args: Extra positional arguments
        **kwargs: Extra keyword arguments
        Returns:
             list of preprocessor names
        """

    def describe(self, identifier: Any, *args, **kwargs) -> dict:
        """
        Describes a preprocessor referred by ``identifier``
        Args:
            identifier: :class:`Preprocessor` or the name of the preprocessor
            **kwargs: Extra keyword arguments
        Returns: Dictionary that describes preprocessor
        """
        preprocessor = identifier
        if isinstance(identifier, str):
            preprocessor = self._create(identifier, *args, **kwargs)
        return json.loads(tamm_json.dumps(preprocessor))

    def __add__(self, other: "PreprocessorRegistry") -> "PreprocessorRegistry":
        """
        Combines two PreprocessorCatalog instances into a composite catalog.

        The preprocessors defined in `self` take precedence over those in `other`
        for the same `model_id`.

        Args:
            other (:obj: `PreprocessorCatalog`): Another preprocessor catalog to combine.

        Returns:
            :obj: `PreprocessorCatalog`: A composite catalog combining `self` and `other`.
        """

        return ComposedPreprocessorRegistry(self, other)

    def __iadd__(self, other: "PreprocessorRegistry") -> "PreprocessorRegistry":
        """
        In-place version of __add__
        """
        return ComposedPreprocessorRegistry(self, other)


class ComposedPreprocessorRegistry(PreprocessorRegistry):
    """
    Represents a composite catalog combining two PreprocessorCatalog instances.
    """

    def __init__(self, *tokenizer_registries: "PreprocessorRegistry"):
        super().__init__()
        self._registries: Sequence["PreprocessorRegistry"] = tokenizer_registries

    def _create(
        self, identifier: Union[str, "Preprocessor"], *args, **kwargs
    ) -> "Preprocessor":
        """
        Creates a preprocessor by identifier.

        Args:
            identifier (:obj: `Union[str, Preprocessor]`): The identifier or preprocessor object.
            **kwargs: Additional keyword arguments for preprocessor creation.

        Returns:
            Preprocessor: The created preprocessor object.
        """
        for registry in self._registries:
            try:
                return registry.create(identifier, *args, **kwargs)
            except Exception as e:
                logger.debug("Cannot create %s with %s: %s", identifier, registry, e)
        raise KeyError(
            f"{identifier} is not found in all available tokenizer registries."
        )

    def list_objects(self, *args, **kwargs) -> List[str]:
        """
        Lists all preprocessor identifiers from both catalogs.

        Args:
            *args: Positional arguments (unused).
            **kwargs: Keyword arguments (unused).

        Returns:
            `List[str]`: A combined list of preprocessor identifiers.
        """
        iterators = [
            registry.list_objects(*args, **kwargs) for registry in self._registries
        ]
        return list(chain(*iterators))

    def clear_cache(self):
        for registry in self._registries:
            registry.clear_cache()



"""
Implements functionality related to auto-classes in the transformers package.
"""

import functools as _functools
import logging as _logging

import transformers as _transformers

from tamm import _helpers
from tamm.hf.transformers.models import _mixins
from tamm.hf.transformers.models import causal_lm as _causal_lm
from tamm.hf.transformers.models import sentencepiece as _sentencepiece

_logger = _logging.getLogger(__name__)


@_helpers.cache  # execute the function only once
def register_with_auto_classes() -> None:
    """
    Register tamm Hugging Face classes with auto classes from :mod:`transformers`,
    such as ``AutoConfig``, ``AutoModel``, and ``AutoModelForCausalLM``.

    Example:

        .. code-block:: python

            register_with_auto_classes()

            from transformers import AutoConfig, AutoModel

            config = AutoConfig.for_model("tamm_causal_lm")
            AutoModel.from_config(config)
    """
    _transformers.AutoConfig.register(
        model_type=_causal_lm.TammCausalLMConfig.model_type,
        config=_causal_lm.TammCausalLMConfig,
    )
    _transformers.AutoModel.register(
        config_class=_causal_lm.TammCausalLMConfig,
        model_class=_causal_lm.TammCausalLMModel,
    )
    _transformers.AutoModelForCausalLM.register(
        config_class=_causal_lm.TammCausalLMConfig,
        model_class=_causal_lm.TammCausalLMForCausalLM,
    )
    _transformers.AutoTokenizer.register(
        config_class=_sentencepiece.TammSentencePieceConfig,
        slow_tokenizer_class=_sentencepiece.TammSentencePieceTokenizer,
    )
    _register_with_auto_classes_deprecated()
    _logger.info("Registered tamm with transformers auto classes")


def _register_with_auto_classes_deprecated():
    """
    Helper function for organizing the registration of deprecated types
    with HF auto classes.
    """


def _auto_register_tamm_classes(fn):
    """
    A decorator that wraps a function in order to call
    :func:`register_with_auto_classes` before calling the function.
    """

    @_functools.wraps(fn)
    def wrapper(*args, **kwargs):
        register_with_auto_classes()
        return fn(*args, **kwargs)

    return wrapper


def _add_tamm_auto_registration_to_auto_class(cls):
    """
    A class decorator that wraps methods of transformers autoclass
    subclasses in order to register tamm objects automatically as needed.
    """

    method_names = ["from_pretrained", "from_config", "for_model"]
    for name in method_names:
        if not hasattr(cls, name):
            continue
        new_fn = _auto_register_tamm_classes(getattr(cls, name))
        setattr(cls, name, new_fn)
    return cls


@_add_tamm_auto_registration_to_auto_class
class AutoConfig(_transformers.AutoConfig):
    """
    Extends :class:`transformers.AutoConfig` to automatically register
    tamm objects when used.
    """


@_add_tamm_auto_registration_to_auto_class
class AutoModel(_transformers.AutoModel, _mixins.FromModelIDMixin):
    """
    Extends :class:`transformers.AutoModel` to add :meth:`from_tamm_model_id` and
    :meth:`save_pretrained_for_tamm_model_id` methods.  This class also
    automatically calls :func:`register_with_auto_classes` as needed.
    """


@_add_tamm_auto_registration_to_auto_class
class AutoModelForCausalLM(
    _transformers.AutoModelForCausalLM, _mixins.FromModelIDMixin
):
    """
    Extends :class:`transformers.AutoModelForCausalLM` to add :meth:`from_tamm_model_id`
    and :meth:`save_pretrained_for_tamm_model_id` methods.  This class also
    automatically calls :func:`register_with_auto_classes` as needed.
    """


@_add_tamm_auto_registration_to_auto_class
class AutoTokenizer(_transformers.AutoTokenizer):
    """
    Extends :class:`transformers.AutoTokenizer` to add
    :meth:`from_tamm_tokenizer_id` and
    :meth:`save_pretrained_for_tamm_tokenizer_id` methods.  This class also
    automatically calls :func:`register_with_auto_classes` as needed.
    """



import warnings as _warnings
from typing import Iterable as _Iterable
from typing import List as _List
from typing import Optional as _Optional
from typing import Union as _Union

import torch as _torch
from torch import nn as _nn

from tamm.context_vars import resolve_device as _resolve_device
from tamm.layers.common._marker import get_marker as _get_marker
from tamm.layers.common.builder import BuildableMixin as _BuildableMixin
from tamm.typing import OptionalDeviceOrString as _OptionalDeviceOrString
from tamm.typing import StateDictType as _StateDictType
from tamm.utils import _is_same_device_type
from tamm.utils import torch_utils as _torch_utils
from tamm.utils._pretrained import fetch_checkpoint


class PretrainedLoader(_BuildableMixin):
    """
    Fetches and loads pretrained state into models.

    Args:
        path (:obj:`str`): A URI to a pretrained ``torch`` checkpoint (saved with
            :func:`torch.save`).
        target_device: The device to place meta tensors.  If ``None``, meta tensors
            end up on the ``torch`` default device.
        keys (iterable of :obj:`str`): An iterable of expected keys in the state
            dict.  Loading raises a :obj:`RuntimeError` if the keys in the checkpoint
            do not match this set.  Defaults to all keys in the module's state dict.
    """

    def __init__(
        self,
        *,
        path: _Union[str, _List[str], None] = None,
        target_device: _OptionalDeviceOrString = None,
        keys: _Optional[_Iterable[str]] = None,
    ):
        self.path = path
        self.target_device = target_device
        self.keys = set(keys) if keys is not None else None

    def load(self, module: _nn.Module) -> None:
        if self.path is None:
            return
        resolved_device = _resolve_device(self.target_device)
        if resolved_device is not None and _is_same_device_type(
            resolved_device, _torch.device("meta")
        ):
            raise RuntimeError(
                "`PretrainedLoader(...).load()` resolved 'meta' as target device which "
                "cannot be loaded with real `nn.Parameter`s "
                "If torch.get_default_device()=='meta', override "
                "'target_device' with a concrete device "
                "such as torch.device('cpu'); Or use a concrete device context"
            )
        if self.keys is None:
            self.keys = set(module.state_dict().keys())

        self._initialize_non_persistent_buffers(module=module, device=resolved_device)

        state_dict = fetch_checkpoint(self.path, map_location=resolved_device)
        result = self._load_state_dict(module=module, state_dict=state_dict)
        self._validate_keys(result)

    def _initialize_non_persistent_buffers(self, module, device):
        for layer in module.modules():
            if _get_marker(layer).weights_initialized:
                continue
            buffers = dict(layer.named_buffers(recurse=False, remove_duplicate=False))
            non_persistent_buffer_names = list(
                buffers.keys() - layer.state_dict().keys()
            )

            if len(non_persistent_buffer_names) == 0:
                continue

            non_persistent_buffers = {
                name: buffers[name] for name in non_persistent_buffer_names
            }

            if any(
                buffer.device.type == "meta"
                for buffer in non_persistent_buffers.values()
            ):
                layer.to_empty(device=device, recurse=False)
                if not hasattr(layer, "reset_parameters"):
                    raise RuntimeError(
                        f"Could not initialize non-persistent buffer, since {layer} has no "
                        "reset_parameters() method"
                    )
                layer.reset_parameters()
            else:
                layer.to(device=device)

    def _validate_keys(
        self, result: "_torch.nn.modules.module._IncompatibleKeys"
    ) -> None:
        """
        Validates all keys in `self.keys` do not present in result.missing_keys
        and logs extra keys from state_dict.
        """

        missing_keys = (
            [key for key in result.missing_keys if key in self.keys]
            if self.keys is not None
            else result.missing_keys
        )
        if missing_keys:
            raise RuntimeError(
                f"Could not load {self.path} into the model because the following "
                f"keys are missing from this checkpoint: {missing_keys}"
            )

        if result.unexpected_keys:
            _warnings.warn(
                f"The checkpoint {self.path} contains unexpected keys "
                f"that have not been loaded into the model: {result.unexpected_keys}"
            )

    def _load_state_dict(self, *, module: _nn.Module, state_dict: _StateDictType):
        """
        Calls module.load_state_dict(state_dict, assign=True) with extra logic to handle
        self.keys and to prevent the dtype of parameters from changing.

        Important: We sometimes override module.load_state_dict() to rename keys on the
        fly.  Thus, the state_dict keys do not necessarily match the parameter names.
        """

        original_dtypes_and_requires_grads = {
            name: (tensor.dtype, tensor.requires_grad)
            for name, tensor in _torch_utils.iter_named_parameters_and_buffers(module)
        }

        ignored_keys = set(module.state_dict().keys()) - self.keys
        preserved_weights = {
            key: value
            for key, value in module.state_dict().items()
            if key in ignored_keys
        }
        result = module.load_state_dict(state_dict, assign=True, strict=False)
        module.load_state_dict(preserved_weights, assign=True, strict=False)

        def update_tensor(name, tensor):
            dtype, requires_grad = original_dtypes_and_requires_grads[name]
            tensor = tensor.type(dtype)
            tensor = tensor.contiguous()
            tensor.requires_grad_(requires_grad)
            return tensor

        _torch_utils.map_named_parameters_and_buffers(
            update_tensor, module, use_new_requires_grad_values=True
        )

        return result



"""
transformer.layer_sequence
^^^^^^^^^^^^^^^^^^^^^^^^^^

This submodule implements the sequence of transformer layers within a
:mod:`transformer stack <tamm.layers.transformer.stack>`.

.. autoclass:: tamm.layers.transformer.BaseTransformerLayerSequence
    :members:

.. autoclass:: tamm.layers.KVReuseTransformerLayerSequence
    :show-inheritance:

.. autoclass:: tamm.layers.TransformerLayerSequence
    :show-inheritance:

.. autoclass:: tamm.layers.UniformTransformerLayerSequence
    :show-inheritance:
"""

from tamm.layers.transformer.layer_sequence.common import BaseTransformerLayerSequence
from tamm.layers.transformer.layer_sequence.kv_reuse import (
    KVReuseTransformerLayerSequence,
)
from tamm.layers.transformer.layer_sequence.uniform import (
    UniformTransformerLayerSequence,
)
from tamm.layers.transformer.layer_sequence.vanilla import TransformerLayerSequence



import contextlib as _contextlib
from typing import ContextManager, Literal, Union

import torch as _torch

from tamm import _helpers
from tamm.layers.common.context_hooks.base import _BaseContextHook


@_contextlib.contextmanager
def dtype_context_manager(
    dtype: Union[
        _torch.dtype, Literal["float", "double", "float16", "float32", "bfloat16"]
    ]
):
    original_dtype = _torch.get_default_dtype()
    dtype = _helpers.get_dtype_from_maybe_string(dtype)
    _torch.set_default_dtype(dtype)
    try:
        yield
    finally:
        _torch.set_default_dtype(original_dtype)


class DtypeContextHook(_BaseContextHook):
    def __init__(self, dtype=None):
        self.dtype = dtype

    def is_null(self) -> bool:
        return self.dtype is None

    def _get_context_manager(self) -> "ContextManager":
        return dtype_context_manager(self.dtype)



import abc
import logging as _logging
import tempfile as _tempfile
from pathlib import Path as _Path
from typing import ContextManager

from tamm import _helpers
from tamm.utils import user_dir_utils as _user_dir_utils
from tamm.utils.uri._helpers import _get_dest_path_from_uri, atomic_download_prefix

_logger = _logging.getLogger(__name__)


class BaseURLHandler(metaclass=abc.ABCMeta):
    name = "base"

    def __init__(self, use_cache=False):
        self.use_cache = use_cache

    def open(self, urlpath: str, /, *args, **kwargs) -> ContextManager:
        _logger.debug("Opening %s using %s", urlpath, self.__class__.__name__)
        local_path = self.map_to_local(urlpath)
        # pylint: disable-next=unspecified-encoding
        return open(local_path, *args, **kwargs)

    def map_to_local(self, uri: str) -> _Path:
        if not isinstance(uri, str):
            raise ValueError(f"uri must be a string, got {type(uri)}")

        return self._map_to_local(uri)

    @abc.abstractmethod
    def _map_to_local(self, uri: str) -> _Path:
        """
        Mapping a remote URI to a local Posix path

        Args:
            uri: source: URI to a file

        Returns: a local Posix path

        """


class AtomicDownloadURIHandler(BaseURLHandler, metaclass=abc.ABCMeta):
    """
    Abstract class for any URI handler that provides "downloading" functionality.
    """

    name = "atomic"

    @abc.abstractmethod
    def _download_uri_to_local_impl(self, uri: str, local_path: _Path) -> _Path:
        """
        Abstract method which downloads a remote URI to a local

        Args:
            uri: URI to be opened by the concrete downloader
            local_path: local path to write output file

        Returns: a local Posix path

        """

    def _download_uri_to_local(self, uri: str, local_path: _Path) -> _Path:
        _logger.debug(
            "Downloading %s to %s using %s", uri, local_path, self.__class__.__name__
        )

        return self._download_uri_to_local_impl(uri, local_path)

    def _map_to_local(self, uri: str) -> _Path:
        """
        Implements downloading with caching and multiprocess safety
        Downloads a file located by ``uri`` to a local path. If cache is enabled,
        the file is expected to be stored under a uri-hashed path under
        get_tamm_cache_dir(). Otherwise, the file is expected to be under
        get_tamm_tmp_dir() with random prefix.

        Args:
            uri: any valid URI

        Returns: a local Posix path

        """
        if _Path(uri).is_file():
            return _Path(uri)
        if uri.startswith("file://"):
            return _Path(uri[len("file://") :])

        dest_path = _get_dest_path_from_uri(uri)
        if self.use_cache and dest_path.is_file():
            _logger.debug(
                "[%s] %s local cache-hit",
                self.__class__.__name__,
                str(dest_path),
            )
            return dest_path

        if not self.use_cache:
            temp_prefix = _Path(
                _tempfile.mkdtemp(dir=_user_dir_utils.get_tamm_tmp_dir())
            )
            temp_prefix.mkdir(parents=True, exist_ok=True)
            temp_path = temp_prefix / dest_path.name
            self._download_uri_to_local(uri, temp_prefix / dest_path.name)
            return temp_path

        with _helpers.file_lock(dest_path.with_suffix(".write_lock")):
            if dest_path.is_file():
                # When a process acquires the lock, first check if the destination path
                # exists. If it exists, return immediately and *avoid*
                # repeated downloading
                return dest_path

            with atomic_download_prefix(dest_path) as temp_prefix:
                self._download_uri_to_local(uri, temp_prefix / dest_path.name)

        return dest_path



import logging as _logging
from typing import Callable as _Callable
from typing import Optional as _Optional
from typing import Union as _Union

from torch import nn as _nn

from tamm import model_repo
from tamm._adapters_v1.utils import maybe_rekey_state_dict
from tamm.model_repo.publishing import PublishedModelConfig as _PublishedModelConfig
from tamm.models.checksum import _checksum
from tamm.models.common import ModelConfig as _ModelConfig

_logger = _logging.getLogger(__name__)


class UnmatchedSourceKeyError(RuntimeError):
    pass


def _get_source_module(
    model_spec: _Union[_nn.Module, _Callable[..., _nn.Module]],
    source_spec: _Optional[
        _Union[_nn.Module, str, _ModelConfig, _PublishedModelConfig]
    ] = None,
    device="cpu",
) -> _nn.Module:
    """
    Get a fully instantiated source model on CPU for hashing
    """
    create_kwargs = {"pretrained": True, "device": device}
    if source_spec is not None:
        if isinstance(source_spec, _nn.Module):
            return source_spec
        if isinstance(source_spec, str):
            # source_spec is model ID
            return model_repo.get_model_repo_lazy().create_model(
                source_spec, **create_kwargs
            )
        if isinstance(source_spec, _ModelConfig):
            return source_spec.create_model(**create_kwargs)
        if isinstance(source_spec, _PublishedModelConfig):
            return source_spec.model_config.create_model(**create_kwargs)
        raise ValueError(
            f"Unknown source_spec {source_spec} " f"(type {type(source_spec)})"
        )
    _logger.debug(
        "source_spec is None, using model metadata to determine the source model"
    )
    if isinstance(source_spec, _nn.Module):
        model = model_spec
    else:
        # model_spec is a Callable
        # create model on meta because we only need metadata
        model = model_spec(pretrained=False, device="meta")
    try:
        source_model_id = model.metadata.source_model_tamm_id
    except AttributeError as e:
        raise AttributeError(
            "source_spec required, if model doesn't have ``source_tamm_model_id``"
        ) from e

    model_config = model_repo.get_model_repo_lazy().create_model_config(source_model_id)
    return model_config.create_model(**create_kwargs)


def compare_checksum_with_source(
    model: _nn.Module,
    source_spec: _Optional[
        _Union[_nn.Module, str, _ModelConfig, _PublishedModelConfig]
    ] = None,
) -> bool:
    # pylint:disable=line-too-long
    """
    Compares ``model`` weights with ``source model``'s weights

    By default, ``model.metadata.source_model_tamm_id`` will
    be used to determine source model weights to compare with.
    Optionally, user can supply ``source_spec``
    to compare source model's weights against weights defined by another
    ``source_model``, ``source_model_config`` or ``source_model_id``

    Args:
        model (:obj:`nn.Module`): Any |tamm| model
        source_spec (:obj:`nn.Module` or model ID :obj:`str` or :obj:`PublishedModelConfig` or :obj:`ModelConfig`, optional):
            alternate source of the model to compare with

    Examples:

        * Compare model's **backbone** with another pretrained-model:

        Suppose you want to compare a LoRA fine-tuned model with its backbone,you
        may use this utility function with an explicit reference to the backbone model
        (i.e., model w/o adapter), like below

        .. code-block:: python

            model = tamm.create_model(model_id)
            compare_checksum_to_source(
                model,
                source_spec=model.metadata.source_model_details["backbone_tamm_model_id"]
            )

        Note that the specification ``source_spec=model.metadata.source_model_details["backbone_tamm_model_id"]`` is required because
        by default, this function uses ``model.metadata.source_model_tamm_id`` as a comparison target.

        .. tip::

            ``source_spec`` can be model config or the model instance if needed. e.g.,

            .. code-block:: python

                compare_checksum_to_source(
                    model,
                    source_spec=tamm.create_model_config(model_id)

                )

            or

            .. code-block:: python

                compare_checksum_to_source(
                    model,
                    source_spec=tamm.create_model(model_id)
                )

        * Compare a model with its original copy:

        Suppose you want to compare a fine-tuned model
        with the original copy from |tamm|, use

        .. code-block:: python

            model = tamm.create_model(model_id)
            # apply changes to the model, such as fine-tuning
            assert not compare_checksum_to_source(
                model
            ) # the output is expected to be False because the model is altered

    Returns:
        ``True`` if ``model``'s weights are the same as the reference
        source, otherwise ``False``

    """

    def model_factory(pretrained, device):  # pylint: disable=unused-argument
        return model

    return compare_factory_checksum_with_source(model_factory, source_spec=source_spec)


def compare_factory_checksum_with_source(
    model_factory: _Callable[..., _nn.Module],
    source_spec: _Optional[
        _Union[_nn.Module, str, _ModelConfig, _PublishedModelConfig]
    ] = None,
) -> bool:
    """
    Compares ``model`` weights with ``source model``'s weights, using model factory
    to defer the creation of candidate model.

    model_factory(...) is expected to return the actual model. This function requires
    half of the peak memory then :meth:`compare_checksum_with_source` because
    ``model`` and ``source model`` are sequentially loaded.

    Args:
        model_factory: A callable which understands argument
            pretrained(=[True, False]) and device(=["meta", "cpu"]).
        source_spec: See :meth:`compare_checksum_with_source`

    .. tip::

        model_factory can be a partially evaluated :meth:`tamm.create_model`

        .. code-block:: python

            from functools import partial

            model_factory = partial(tamm.create_model, model_id)
            assert compare_checksum_with_source_factory(model_factory)

    Returns:
         ``True`` if ``model``'s weights are the same as the reference
         source, otherwise ``False``
    """
    # pylint:enable=line-too-long
    source_model = _get_source_module(model_factory, source_spec)
    source_sd = source_model.state_dict()
    source_keys = frozenset(source_sd.keys())

    def _compute_checksum(_model, name):
        """
        Only compute checksum within the subset of state dictionary filtered by
        ``source_keys``
        """
        _logger.debug("model has keys:")
        for key in sorted(list(_model.keys())):
            _logger.debug("\t%s", key)
        _logger.debug("Computing checksum for %s on keys:", name)
        for key in sorted(list(source_keys)):
            _logger.debug("\t%s", key)
        return _checksum.state_dict_checksum(_model, source_keys)

    source_checksum = _compute_checksum(source_sd, "source_model")
    _logger.debug("source_checksum = %s", source_checksum)
    del source_sd  # free memory once source model has been hashed
    candidate_model = model_factory(pretrained=True, device="cpu")
    candidate_state_dict = candidate_model.state_dict()
    candidate_state_dict = maybe_rekey_state_dict(source_model, candidate_state_dict)
    del source_model  # free memory once source model has been hashed
    unmatched_source_keys = source_keys - frozenset(candidate_state_dict.keys())
    if unmatched_source_keys:
        raise UnmatchedSourceKeyError(
            f"source model has the following keys which don't "
            f"exist in 'model': "
            f"'{unmatched_source_keys}'"
        )

    model_checksum = _compute_checksum(candidate_state_dict, "model")
    _logger.debug("model_checksum = %s", model_checksum)
    return model_checksum == source_checksum



from tamm.converters.models.common import ModelStateDictConverter
from tamm.converters.models.from_tamm import convert_from_tamm_state_dict, save
from tamm.converters.models.registry import list_converters
from tamm.converters.models.to_tamm import convert_to_tamm_state_dict, load



from tamm.hf.transformers.models.causal_lm.configuration_causal_lm import (
    TammCausalLMConfig,
)
from tamm.hf.transformers.models.causal_lm.modeling_causal_lm import (
    TammCausalLMForCausalLM,
    TammCausalLMModel,
    TammCausalLMPreTrainedModel,
)



"""
This module provides variations of dropout layers.
"""

import torch.nn as _nn

from tamm.layers.common import LayerMixin as _LayerMixin


class Dropout(_nn.Dropout, _LayerMixin):
    """A dropout layer with a default dropout rate of zero."""

    def __init__(self, p: float = 0.0, inplace: bool = False) -> None:
        super().__init__(p=p, inplace=inplace)

    def forward(self, *args, **kwargs):  # pylint: disable=useless-parent-delegation
        # This override is important for torch.compile with DDP as of torch 2.5
        return super().forward(*args, **kwargs)



"""
transformer.token_type_encoding
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

This module implements token type encoding layers for :class:`.TransformerStack`.
These layers encode token types into embeddings and attention masks prior to the
sequence of transformer layers.

.. autoclass:: tamm.layers.TransformerTokenTypeEncoding
    :show-inheritance:
    :members: forward

.. autoclass:: TransformerTokenTypeEncodingOutput

.. autoclass:: tamm.layers.TransformerTokenTypeEmbedding
    :show-inheritance:
    :members: create_builder
"""


import abc as _abc
import collections as _collections
from typing import Any as _Any
from typing import Dict as _Dict
from typing import Optional as _Optional

import torch as _torch
import torch.nn as _nn

from tamm import _helpers
from tamm.layers import embedding as _embedding
from tamm.layers.common import ConfigurableLayerMixin as _ConfigurableLayerMixin
from tamm.layers.transformer import token_metadata as _token_metadata
from tamm.typing import ModuleBuilder as _ModuleBuilder

TransformerTokenTypeEncodingOutput = _collections.namedtuple(
    "TransformerTokenTypeEncodingOutput", ["embeddings", "attention_side_inputs"]
)
"""
A :class:`namedtuple` for holding outputs from a :class:`TransformerTokenTypeEncoding`.

.. py:attribute:: embeddings

    Embedding inputs to the first layer of a transformer.

.. py:attribute:: attention_side_inputs

    Side inputs to self attention layers of a transformer.
"""


class TransformerTokenTypeEncoding(_nn.Module, _abc.ABC):
    """
    A base class for layers that encode token types into the inputs of
    a transformer layer sequence.
    """

    def __init__(self):
        super().__init__()

    @_abc.abstractmethod
    def forward(
        self,
        *,
        token_metadata: _token_metadata.TokenMetadata,
        embeddings: _torch.Tensor,
        attention_side_inputs: _Dict[_Any, _Any],
    ) -> TransformerTokenTypeEncodingOutput:
        """
        Args:
            token_metadata (:obj:`TokenMetaData`): A :obj:`TokenMetaData`
                object, which includes a ``q_token_types`` tensor with shape
                ``(batch_size, seq_len)``.
            embeddings (:obj:`torch.Tensor`): The input embeddings to a transformer
                model with shape ``(batch_size, seq_len, hidden_dim)``.  The dtype
                should match the computation dtype for attention layers.
            attention_side_inputs (:obj:`dict`): A dictionary of side inputs to
                attention layers.  This should contain ``"attention_mask"`` and
                ``"flash_attention_options"`` keys.

        Returns:
            A :obj:`TransformerTokenTypeEncodingOutput`, which contains new
            ``embeddings`` and ``attention_side_inputs`` objects with
            token types encoded.
        """


class TransformerTokenTypeEmbedding(
    TransformerTokenTypeEncoding, _ConfigurableLayerMixin
):
    """
    This layer maps token types to trainable embedding tensors and then adds these
    embeddings to the token embeddings.

    Args:
        embedding: An embedding layer (or builder) that maps token types to embeddings.
        sequence_start (:obj:`int`, optional): An option that controls which tokens
            receive token type embeddings.  If ``sequence_start > 0``, then the first
            ``sequence_start`` tokens do not receive them.
    """

    # pylint: disable=duplicate-code

    def __init__(
        self,
        embedding: _ModuleBuilder,
        sequence_start: _Optional[int] = None,
    ):
        super().__init__()
        self.embedding = _helpers.maybe_build_module(embedding)
        self.sequence_start = sequence_start

    @classmethod
    def create_basic_builder(  # pylint: disable=arguments-differ
        cls,
        num_embeddings: int,
        embedding_dim: int,
        *,
        sequence_start: _Optional[int] = None,
    ) -> _ModuleBuilder:
        """
        Creates and returns a builder for :class:`TransformerTokenTypeEmbedding` layers.

        Args:
            num_embeddings (:obj:`int`): The number of embeddings in the token type
                embedding table.
            embedding_dim (:obj:`int`): The dimension of the embeddings.
            sequence_start (:obj:`int`, optional): The ``sequence_start`` option for the
                layer.

        Returns:
            The configured :obj:`LayerBuilder`.
        """
        embedding = _embedding.Embedding.Builder(
            num_embeddings=num_embeddings,
            embedding_dim=embedding_dim,
        )
        return cls.Builder(
            embedding=embedding,
            sequence_start=sequence_start,
        )

    def forward(  # pylint: disable=arguments-differ
        self,
        *,
        token_metadata: _token_metadata.TokenMetadata,
        embeddings: _torch.Tensor,
        attention_side_inputs: _Dict[_Any, _Any],
    ) -> TransformerTokenTypeEncodingOutput:
        token_types = token_metadata.q_token_types

        if self.sequence_start in (None, 0):
            start_embeddings = None
        else:
            start_embeddings = embeddings[:, : self.sequence_start]
            embeddings = embeddings[:, self.sequence_start :]
            token_types = token_types[:, self.sequence_start :]

        type_embeddings = self.embedding(token_types)
        type_embeddings = type_embeddings.type_as(embeddings)
        outputs = embeddings + type_embeddings

        if start_embeddings is not None:
            outputs = _torch.cat([start_embeddings, outputs], dim=1)

        return TransformerTokenTypeEncodingOutput(
            embeddings=outputs, attention_side_inputs=attention_side_inputs
        )

    def extra_repr(self) -> str:
        if self.sequence_start in (None, 0):
            return ""
        return f"sequence_start={self.sequence_start}"



import sys

from tamm._cli.common import HEADER, get_extended_root_parser


def main(args=None):
    parser = get_extended_root_parser()

    args = args if args is not None else sys.argv[1:]
    if len(args) == 0:
        print(HEADER)
        parser.print_help()
        return

    parsed_args = parser.parse_args(args)
    kwargs = vars(parsed_args)
    kwargs = {k.replace("-", "_"): v for k, v in kwargs.items()}
    if "func" in kwargs:
        func = kwargs.pop("func")
        func(**kwargs)
    else:
        parser.parse_args([*args, "--help"])


if __name__ == "__main__":
    main()



from typing import List, Union

from tamm.tokenizers.common import Tokenizer, TokenizerConfig
from tamm.tokenizers.registry import TokenizerRegistry


class PlaceholderTokenizerRegistry(TokenizerRegistry):
    def _create_config(
        self, identifier: Union[str, "Tokenizer"], *args, **kwargs
    ) -> "TokenizerConfig":
        raise KeyError(f"{identifier} is not found.")

    def list_(self, *args, **kwargs) -> List[str]:
        return []

    def describe(self, identifier: str, *args, **kwargs) -> str:
        raise KeyError(f"{identifier} is not found, cannot describe.")

    def __iadd__(self, other):
        return other

    def __add__(self, other):
        return other



from typing import Any as _Any
from typing import Callable as _Callable
from typing import Dict as _Dict
from typing import List as _List
from typing import Optional as _Optional

from tamm._plugin import import_discovered_plugins
from tamm.utils import registry as _registry

_REGISTRY = _registry.Registry("Model state dict converters")


def register_converter_cls(
    converter_cls, converter_id: str, description: _Optional[str] = None
):
    _REGISTRY.register(converter_cls, key=converter_id, description=description)


def get_converter_cls(converter_id: str) -> _Callable:
    """
    Retrieve the registered converter class from the registry based on the provided
    converter ID. Also discovers the converters from all installed plugins.
    """
    import_discovered_plugins()
    return _REGISTRY.get_factory_fn(converter_id)


def create_converter_from_tamm_state_dict(
    converter_id: str, state_dict: _Dict[str, _Any]
):
    converter_cls = get_converter_cls(converter_id)
    return converter_cls.from_tamm_state_dict(state_dict)


def create_converter_from_other_state_dict(
    converter_id: str, state_dict: _Dict[str, _Any]
):
    converter_cls = get_converter_cls(converter_id)
    return converter_cls.from_other_state_dict(state_dict)


def list_converters(
    include_descriptions: bool = False,
    filter_deprecated: _Optional[bool] = True,
) -> _List[str]:
    """
    List the registered converter IDs from all installed plugins.
    """
    import_discovered_plugins()  # Ensure all plugins are loaded before listing
    return _REGISTRY.list(
        include_descriptions=include_descriptions,
        filter_deprecated=filter_deprecated,
    )



"""Command for converting model checkpoints from tamm to other formats."""

import collections as _collections
import logging
from typing import Optional

from tamm import _helpers, converters
from tamm._cli.common import argument, command

logger = logging.getLogger(__name__)


def parse_extra_metadata(metadata: str) -> dict:
    """Helper method to convert a metadata string to a dictionary,
    metadata string should follow the format "key1=val1,key2=val2".
    """
    expected_keys = {"num_heads": int, "num_kv_heads": int}

    metadata = dict([item.split("=") for item in metadata.split(",")])
    results = {}
    for key, cast_type in expected_keys.items():
        val = metadata.pop(key, None)
        if val:
            results[key] = cast_type(val)
    return results


@command("from-tamm", help_text="Convert model checkpoints from tamm to a new format")
@argument(
    "--metadata",
    type=str,
    help="Optional metadata for the state dict in the format key1=val1,key2=val2",
)
@argument(
    "--dtype",
    type=str,
    help="Optional output dtype, such as float16, float32, etc",
)
@argument("dest", type=str, help="Destination path for the converted model state")
@argument(
    "source",
    type=str,
    help="Source path to a tamm model state dict (saved by :func:`torch.save`)",
)
@argument(
    "converter-id",
    type=str,
    help=(
        "A converter id that specifies how to do the conversion--use "
        "``tamm ls converters -l`` for descriptions of available converters"
    ),
)
def from_tamm(
    converter_id: str,
    source: str,
    dest: str,
    dtype: Optional[str] = None,
    metadata: str = None,
) -> None:
    logger.info(f"Initiating conversion: Loading tamm state dict from {source}.")
    state_dict = _helpers.safe_torch_load(source)

    if metadata is not None:
        metadata = parse_extra_metadata(metadata)
        if len(metadata) > 0:
            state_dict = _collections.OrderedDict(state_dict)
            # pylint: disable=protected-access
            metadata.update(
                state_dict._metadata if hasattr(state_dict, "_metadata") else {}
            )
            state_dict._metadata = metadata
            logger.info(f"Added extra metadata {metadata} to state dict.")

    converters.save(state_dict, dest, converter=converter_id, dtype=dtype)
    logger.info(f"Conversion completed: Model state saved to {dest}.")



from typing import ContextManager, Optional

from tamm.context_vars import model_build_device_context as _model_build_device_context
from tamm.context_vars import resolve_pretrained_flag as _resolve_pretrained_flag
from tamm.layers.common.context_hooks.base import _BaseContextHook
from tamm.typing import LenientOptionalBool
from tamm.utils import OptionalBool as _OptionalBool


class UseMetaInitTrickContextHook(_BaseContextHook):
    def __init__(
        self,
        pretrained: "LenientOptionalBool" = _OptionalBool.NOTSET,
        pretrained_path: "Optional[str]" = None,
    ):
        self.pretrained = pretrained
        self.pretrained_path = pretrained_path

    @property
    def should_use_meta_init_trick(self) -> bool:
        return (
            _resolve_pretrained_flag(self.pretrained) == _OptionalBool.TRUE
            and self.pretrained_path is not None
        )

    def is_null(self) -> bool:
        return not self.should_use_meta_init_trick

    def _get_context_manager(self) -> "ContextManager":
        return _model_build_device_context(enable_meta_init_trick=True)



import dataclasses as _dataclasses
from typing import List as _List
from typing import Union as _Union


@_dataclasses.dataclass
class InstructLMPreprocessorOutput:
    """A return type for instruct LM preprocessors in |tamm|."""

    input_ids: _List[int]
    """A list of tokenized prompt ids."""

    label_ids: _Union[_List[int], None]
    """
    An optional list of label ids.  This may be ``None`` during inference.
    """



"""
This module dynamically aggregates model config classes so that we can document them
with Sphinx autodoc.  We do not recommend using this module to access the config
classes (use ``model_cls.Config`` instead).
"""

from tamm import models as _models


def _attach_model_configs_to_module():
    model_objects = (getattr(_models, obj_name) for obj_name in dir(_models))
    config_classes = (
        getattr(model_cls, "Config")
        for model_cls in model_objects
        if hasattr(model_cls, "Config")
    )
    for cls in config_classes:
        globals()[cls.__name__] = cls


_attach_model_configs_to_module()



"""
Publishing
==========

.. autoclass:: tamm.model_repo.publishing.ModelPublisher
.. autoclass:: tamm.model_repo.publishing.PublishedModelConfig

"""

from tamm.model_repo.publishing.base import ModelPublisher
from tamm.model_repo.publishing.config import PublishedModelConfig

__all__ = ["PublishedModelConfig", "ModelPublisher"]



from tamm._cli._listers._base import BaseLister
from tamm._cli._listers._helpers import (
    PrettyPrintListedEntities as _PrettyPrintListedEntities,
)
from tamm.converters import list_converters


class ConverterLister(BaseLister):
    name = "converters"

    # pylint: disable=redefined-builtin
    def __init__(self, all, long, wide, show_deprecated):
        super().__init__(all=all, long=long, wide=wide, show_deprecated=show_deprecated)
        self._listing_function = list_converters

    def print(self):
        _PrettyPrintListedEntities(self._listing_function).print(
            include_descriptions=self.long,
            filter_deprecated=not self.all,
            wide=self.wide,
        )



from typing import ContextManager

from tamm.context_vars import freeze_params_flag_context
from tamm.layers.common.context_hooks.base import _BaseContextHook
from tamm.typing import LenientOptionalBool
from tamm.utils import OptionalBool as _OptionalBool


class FreezeParamsContextHook(_BaseContextHook):
    """
    Applies freeze params context if explicitly provided in
    ``create_builder(freeze_params=[True|False])`` calls.
    Context is not resolved here but deferred to :class:`FreezeParamsPostHook`
    """

    def __init__(self, freeze_params: "LenientOptionalBool" = _OptionalBool.NOTSET):
        self.freeze_params = _OptionalBool(freeze_params)

    def is_null(self) -> bool:
        return self.freeze_params == _OptionalBool.NOTSET

    def _get_context_manager(self) -> "ContextManager":
        return freeze_params_flag_context(self.freeze_params)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.freeze_params})"



"""
This module provides helpers for extending the :mod:`json` module for |tamm|.  Please
see the unit tests for usage examples.
"""

import abc
import logging
from collections import defaultdict
from typing import Any, Callable, Dict
from typing import List as _List
from typing import Mapping, Protocol, Union

import torch

from tamm._plugin.utils import import_named_plugin
from tamm.runtime_configuration import rc as _rc
from tamm.utils import registry as _registry
from tamm.utils.json._hooks import (
    PackageRequirementsHook,
    ReadabilityHook,
    UnusedAttributeHook,
)

logger = logging.getLogger(__name__)

JSONRegistries: Mapping[str, _registry.Registry] = defaultdict(_registry.Registry)


def global_default(obj):
    try:
        # pylint: disable=protected-access
        return obj._to_json_dict()
    except AttributeError as exc:
        raise TypeError from exc


def global_object_hook(obj: dict):
    if not isinstance(obj, dict):
        return obj
    tamm_type_info = obj.get("__tamm_type__", None)
    if not tamm_type_info:
        return obj
    namespace, tamm_type = tamm_type_info.rsplit(":", maxsplit=1)

    if namespace not in JSONRegistries:
        plugin_required = namespace.split(":", maxsplit=1)[0]
        if plugin_required != _rc.PROJECT_SLUG:
            import_named_plugin(plugin_required)

    registry = JSONRegistries[namespace]
    cls = registry.get_factory_fn(tamm_type)
    # pylint: disable=protected-access
    try:
        return cls._from_json_dict(**obj)
    except TypeError as e:
        module_name = tamm_type_info.split(":", maxsplit=1)[0]
        raise TypeError(
            f"Error encountered while deserializing a '{tamm_type_info}': {e}. "
            "Either (1) the object's JSON representation changed in an "
            "unrecognizable way, (2) the deserialization requires a newer "
            f"version of the {module_name} module, or (3) something unexpected "
            "went wrong."
        ) from e


def _resolve_json_namespace(
    bases, explicit_namespace=None, default_namespace="default"
) -> str:
    """
    Helper method to resolve json namespace from multiple inheritance
    Args:
        bases: Base classes which may or may not have _JSON_NAMESPACE attribute
        explicit_namespace: Explicit namespace provided by class Signature
        default_namespace: Default namespace

    Returns: Resolved JSON namespace or default namespace

    """
    if explicit_namespace is not None:
        return explicit_namespace
    resolved_namespace = None
    for base in bases:
        try:
            # pylint: disable=protected-access
            if base._JSON_NAMESPACE not in {
                None,
                default_namespace,
            }:
                resolved_namespace = base._JSON_NAMESPACE
        except AttributeError:
            pass
    if resolved_namespace is None:
        return default_namespace
    return resolved_namespace


def _get_module_prefix(class_name: str):
    """Helper function to return the prefix for a base class."""
    return class_name.split(".", maxsplit=1)[0]


def _retrieve_prefixed_namespace(
    module_prefix: str, resolved_namespace: str, delimiter=":"
):
    """
    Helper function to return the prefixed full namespace.
    Format is of type module_prefix:entity.

    Example:

    .. code-block:: python

        namespace = _retrieve_prefixed_namespace(
            module_prefix="tamm",
            resolved_namespace="models"
        )
        print(namespace)
        "tamm:models"
    """
    return module_prefix + delimiter + resolved_namespace.split(delimiter)[-1]


class JSONSerializableMeta(type):
    def __new__(mcs, clsname, bases, attrs, /, json_namespace=None, **kwargs):
        newclass = super(JSONSerializableMeta, mcs).__new__(
            mcs, clsname, bases, attrs, **kwargs
        )

        module_prefix = _get_module_prefix(attrs["__module__"])

        resolved_namespace = _resolve_json_namespace(bases, json_namespace)

        # pylint:disable=invalid-name
        newclass._JSON_NAMESPACE = _retrieve_prefixed_namespace(
            module_prefix, resolved_namespace
        )

        # pylint:enable=invalid-name
        JSONRegistries[newclass._JSON_NAMESPACE].register(newclass, clsname)
        JSONRegistries[newclass._JSON_NAMESPACE]._name = newclass._JSON_NAMESPACE
        return newclass


class JSONSerializableABCMeta(abc.ABCMeta, JSONSerializableMeta):
    # pylint: disable=line-too-long
    """
    Helper `metaclass <https://docs.python.org/3/reference/datamodel.html#metaclasses>`_
    for classes who inherit from both ``abc.ABC`` and
    ``JSONSerializableMixin``.

    For example,

    .. code-block:: python

        class MyClass(abc.ABC, JSONSerializableMixin): # WRONG, will raise TypeError

    the class signature *has* to be

    .. code-block:: python

        class MyClass(abc.ABC, JSONSerializableMixin, metaclass=JSONSerializableABCMeta):

    instead to avoid

        TypeError: metaclass conflict: the metaclass of a derived class must
        be a (non-strict) subclass of the metaclasses of all its bases

    Reference:
    `StackOverflow <https://stackoverflow.com/questions/11276037/resolving-metaclass-conflicts>`_
    """
    # pylint: enable=line-too-long


class JSONSerializableMixin(metaclass=JSONSerializableMeta):
    """
    A mixin to make ``tamm`` objects JSON serializable.
    """

    _JSON_NAMESPACE: str

    def _package_requirements(self) -> _List[str]:
        """
        Children can provide a package requirements constraint to determine

        Returns:
            a ``list`` of package requirements in
            `PEP508 <https://peps.python.org/pep-0508/>`_
        """
        return []

    @classmethod
    def _from_json_dict_impl(cls, **raw_dict: dict):
        """
        Children overridable classmethod to instantiate an object of this class from
        raw JSON dictionary. Default unpack all attributes to class __init__()

        Args:
            **raw_dict: dictionary representation with all children nodes deserialized

        Returns:
            Deserialized object of this class, *i.e.,* ``cls(**raw_dict)``

        """
        return cls(**raw_dict)

    def _to_json_dict_impl(self) -> dict:
        """
        Children overridable instance method to generate a dictionary representation of
        object.

        .. note::

            Default implementation force cast the object to dict.
            i.e., assumes ``self`` conforms to
            `mapping protocol <https://docs.python.org/3/c-api/mapping.html>`_.

        Returns:
            (:obj:`dict`) JSON dictionary representation of this object

        """
        return dict(self)  # type: ignore[call-overload]

    @classmethod
    def _from_json_dict(cls, **raw_dict: dict):
        for hook in [PackageRequirementsHook(), UnusedAttributeHook()]:
            raw_dict = hook(raw_dict)

        return cls._from_json_dict_impl(**raw_dict)

    def _to_json_dict(self) -> dict:
        raw_dict = self._to_json_dict_impl()
        raw_dict["__tamm_type__"] = f"{self._JSON_NAMESPACE}:{self.__class__.__name__}"
        if self._package_requirements():
            raw_dict["__package_requirements__"] = self._package_requirements()
        for hook in [ReadabilityHook()]:
            raw_dict = hook(raw_dict)
        return raw_dict


class JSONSerializableABCMixin(
    abc.ABC, JSONSerializableMixin, metaclass=JSONSerializableABCMeta
):
    """
    A mixin to make ``tamm`` objects both JSON serializable and an :class:`abc.ABC`.
    Inheriting from this class avoids the :obj:`TypeError` that occurs when inheriting
    from both :class:`ABC` and :class:`JSONSerializableMixin` without specifying the
    :class:`JSONSerializableABCMeta` metaclass.
    """


class JSONHook(Protocol):
    """
    This class defines a Protocol for serializing/desrializing custom types with JSON.
    """

    @abc.abstractmethod
    def default(self, obj: Any) -> Any:
        """
        This method may be called by :func:`json.dumps` for any ``obj`` that is not
        JSON-serializable.  The method should return a JSON-serializable :obj:`dict`
        representing ``obj`` or raise a ``TypeError`` if it is unable to do so.  See
        :func:`json.dumps` for more info.
        """

    @classmethod
    @abc.abstractmethod
    def object_hook(cls, obj: Dict[str, Any]) -> Any:
        """
        This method is called by :func:`json.loads` as an object hook, where ``obj`` is
        the result of any object literal decode.  This method should either (1) perform
        the inverse of :meth:`to_json_serializable` and return a new object represented
        by ``obj``, or (2) return ``obj`` unmodified if it is unrecognized.  See
        :func:`json.loads` for more info.
        """


class DeviceJSONHook(JSONHook):
    """
    This class implements a JSON hook for serializing :class:`torch.device` objects.
    """

    def default(self, obj: Any) -> Any:
        if not isinstance(obj, torch.device):
            raise TypeError(f"cannot serialize {obj} which is not a torch.device")
        return {"__tamm_type__": "torch.device", "type": obj.type, "index": obj.index}

    @classmethod
    def object_hook(cls, obj: Dict[str, Any]) -> Union[torch.device, Dict[str, Any]]:
        if not isinstance(obj, dict):
            return obj
        if obj.get("__tamm_type__") == "torch.device":
            index = obj.get("index", None)
            if index is not None:
                return torch.device(f"{obj['type']}:{index}")
            return torch.device(f"{obj['type']}")
        return obj


class DTypeJSONHook(JSONHook):
    """
    This class implements a JSON hook for serializing :class:`torch.dtype` objects.
    """

    def default(self, obj: Any) -> Any:
        if not isinstance(obj, torch.dtype):
            raise TypeError(f"cannot serialize {obj} which is not a torch.dtype")
        name = str(obj).rsplit(".", maxsplit=1)[-1]
        return {"__tamm_type__": "torch.dtype", "name": name}

    @classmethod
    def object_hook(cls, obj: Dict[str, Any]) -> Union[torch.dtype, Dict[str, Any]]:
        if not isinstance(obj, dict):
            return obj
        if obj.get("__tamm_type__") == "torch.dtype":
            return getattr(torch, obj["name"])
        return obj


class CompositeObjectHooks:
    """
    This class combines multiple ``object_hooks`` arguments for :func:`json.loads`.

    Args:
        *callables (:obj:`callable`):  An arbitrary number of
            callable ``object_hooks`` functions for
            :func:`json.loads`.
    """

    def __init__(
        self,
        *callables: Callable[..., Any],
    ):
        self.callables = callables

    def __call__(self, obj: Any) -> Any:
        """
        Calls all default functions in sequence, returning the result of the first
        call that does not raise a ``TypeError``.  Raises a ``TypeError`` if every
        default raises a ``TypeError``.
        """
        for _callable in self.callables:
            obj = _callable(obj)
        return obj


class CompositeDefault:
    """
    This class combines multiple ``default`` arguments for :func:`json.dumps`.

    Args:
        *callables (:obj:`JSONHook` or :obj:`callable`):  An arbitrary number of
            :obj:`JSONHook` instances or callable ``default`` functions for
            :func:`json.dumps`.
    """

    def __init__(
        self,
        *callables: Callable[..., Any],
    ):
        self.callables = callables

    # pylint: disable-next=inconsistent-return-statements
    def __call__(self, obj: Any) -> Any:
        """
        Calls all default functions in sequence, returning the result of the first
        call that does not raise a ``TypeError``.  Raises a ``TypeError`` if every
        default raises a ``TypeError``.
        """

        for _callable in self.callables:
            try:
                return _callable(obj)
            except TypeError as e:
                logger.debug(
                    "'%s' cannot serialize '%s' due to: '%s'",
                    _callable.__name__,
                    obj.__class__.__name__,
                    e,
                )
        _callables_qualname = ", ".join(
            f"{_callable.__qualname__}()" for _callable in self.callables
        )
        raise TypeError(
            f"Object of type '{type(obj).__name__}' is not "
            f"JSON serializable by CompositeDefaults with callables"
            f" ({_callables_qualname})"
        )
